import logging
from enum import Enum

from typing import List, Callable, Any, Type, Union

from dataprep.parse.model.metadata import PreprocessingMetadata
from dataprep.parse.model.placeholders import placeholders
from dataprep.subtokens import FullTokenIterator, is_terminal_subtoken, SubtokenIterator

logger = logging.getLogger(__name__)

MetricName = str


class TokenTypes(str, Enum):
    ALL = 'all'
    ONLY_COMMENTS = 'only_comments'
    ALL_BUT_COMMENTS = 'all_but_comments'

    def __str__(self) -> str:
        return self.value


def to_full_token_string(subtokens: List[str], include_debug_tokens: bool = False):
    """
    >>> to_full_token_string(['the', 're</t>'], include_debug_tokens=True)
    'the|re</t>'

    >>> to_full_token_string(['re', 'vol', 'v', 'er</t>'])
    'revolver'

    >>> to_full_token_string([placeholders['olc_end']])
    '<EOL>'
    """
    if not isinstance(subtokens, List):
        raise TypeError(f'This methods accept a list of strings, however {type(subtokens)} was passed')

    if len(subtokens) == 1 and subtokens[0] in placeholders.values():
        return subtokens[0]

    sep = '|' if include_debug_tokens else ''
    joined = sep.join(subtokens)
    cwe = placeholders['compound_word_end']
    if not is_terminal_subtoken(joined):
        raise ValueError(f'{joined} ({subtokens}) is not a full token')
    return joined if include_debug_tokens else joined[:-len(cwe)]


TokenIterator = Union[SubtokenIterator, FullTokenIterator]


class FilteringTokenIterator(object):
    """
    >>> from langmodels.evaluation.metrics import inclusion_func_dict
    >>> metadata = PreprocessingMetadata(word_boundaries=[0, 1, 3], comments=[(1,3)])
    >>> [token for token in FilteringTokenIterator(['hi', '/', '/'], metadata, inclusion_function=inclusion_func_dict[TokenTypes.ALL_BUT_COMMENTS])]
    ['hi']

    >>> from langmodels.evaluation.metrics import inclusion_func_dict
    >>> metadata = PreprocessingMetadata(word_boundaries=[0, 1, 3], comments=[(1,3)])
    >>> [token for token in FilteringTokenIterator(['hi', '/', '/'], metadata)]
    ['hi', '//']

    >>> from langmodels.evaluation.metrics import inclusion_func_dict
    >>> metadata = PreprocessingMetadata(word_boundaries=[0, 1, 3], comments=[(1,3)])
    >>> it = FilteringTokenIterator(['hi', '/', '/'], metadata, token_iterator_type=SubtokenIterator, agg=lambda l: l[0])
    >>> [token for token in it]
    ['hi', '/', '/']

    >>> from langmodels.evaluation.metrics import inclusion_func_dict
    >>> metadata = PreprocessingMetadata(word_boundaries=[0, 1, 3], comments=[(1,3)])
    >>> it = FilteringTokenIterator(['hi', '/', '/'], metadata, inclusion_function=inclusion_func_dict[TokenTypes.ONLY_COMMENTS], token_iterator_type=SubtokenIterator, agg=lambda l: l[0])
    >>> [token for token in it]
    ['/', '/']
    """
    def __init__(self, subwords: List[Any], metadata: PreprocessingMetadata,
                 inclusion_function: Callable[[PreprocessingMetadata, int], bool] = lambda m, i: True,
                 token_iterator_type: Type[TokenIterator] = FullTokenIterator,
                 agg: Callable[[List[str]], Any] = lambda s: ''.join(s)):
        self.full_token_iterator = token_iterator_type(subwords, metadata.word_boundaries, agg=lambda s: (s, len(s)))
        self.inclusion_function = inclusion_function
        self.metadata = metadata
        self.agg = agg
        self.current_index = 0

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            next_full_token, n_subtokens = next(self.full_token_iterator)
            current_index = self.current_index
            self.current_index += n_subtokens
            if self.inclusion_function(self.metadata, current_index):
                return self.agg(next_full_token)


from langmodels.evaluation.evaluation import evaluate_model_on_file, evaluate_model_on_string, evaluate_model_on_path, \
    evaluate_model_on_project_set


__all__ = [
    'evaluate_model_on_file',
    'evaluate_model_on_string',
    'evaluate_model_on_path',
    'evaluate_model_on_project_set'
]



import os

import psutil
import time
from collections import defaultdict

from typing import DefaultDict, Dict, List


class TimeMeasurer(object):
    def __init__(self):
        self._times: DefaultDict[str, List[float]] = defaultdict(list)
        self._in_progress: Dict[str, float] = {}

    def tick(self, what: str):
        if what in self._in_progress:
            raise ValueError(f"Tick has been already called for: {what}")

        self._in_progress[what] = time.perf_counter()

    def tock(self, what: str):
        if what not in self._in_progress:
            raise ValueError(f"Tick has not been called for: {what}")

        self._times[what].append(time.perf_counter() - self._in_progress[what])
        del self._in_progress[what]

    def totals(self):
        return {k: sum(v) for k, v in self._times.items()}


def get_cpu_memory_used_mb():
    process = psutil.Process(os.getpid())
    return process.memory_info().rss / 2 ** 20


import sys

import dataprep.api.corpus as api

from langmodels.lmconfig.datamodel import LMTrainingConfig, LstmArch, TrainingProcedure, \
    PrepFunction, Corpus, CosineLRSchedule, PrepFunctionOptions
from langmodels.lmconfig.serialization import load_config_from_string, dump_config_to_string


def test_serialize_deserialize_defaults():
    lm_training_config = LMTrainingConfig(
        base_model=None,
        corpus=Corpus(path='/path/to/corpus', extensions="java"),
        prep_function=PrepFunction(
            api.bpe,
            ['10k'],
            PrepFunctionOptions(no_str=True, no_com=True, no_spaces=False, no_unicode=False, max_str_length=sys.maxsize)
        ),
        arch=LstmArch(), bs=32, bptt=200,
        training_procedure=TrainingProcedure(schedule=CosineLRSchedule(cyc_len=3, max_epochs=30, max_lr=1e-4)))

    assert lm_training_config == load_config_from_string(dump_config_to_string(lm_training_config))



from collections import OrderedDict


def weights_to_v1(old_weights: OrderedDict):
    new_weights = OrderedDict()
    new_weights['0.encoder.weight']=old_weights['0.encoder.weight']
    new_weights['0.encoder_dp.emb.weight']=old_weights['0.encoder_with_dropout.embed.weight']
    new_weights['0.rnns.0.weight_hh_l0_raw']=old_weights['0.rnns.0.module.weight_hh_l0_raw']
    new_weights['0.rnns.0.module.weight_ih_l0']=old_weights['0.rnns.0.module.weight_ih_l0']
    new_weights['0.rnns.0.module.weight_hh_l0']=old_weights['0.rnns.0.module.weight_hh_l0_raw']
    new_weights['0.rnns.0.module.bias_ih_l0']=old_weights['0.rnns.0.module.bias_ih_l0']
    new_weights['0.rnns.0.module.bias_hh_l0']=old_weights['0.rnns.0.module.bias_hh_l0']
    new_weights['0.rnns.1.weight_hh_l0_raw']=old_weights['0.rnns.1.module.weight_hh_l0_raw']
    new_weights['0.rnns.1.module.weight_ih_l0']=old_weights['0.rnns.1.module.weight_ih_l0']
    new_weights['0.rnns.1.module.weight_hh_l0']=old_weights['0.rnns.1.module.weight_hh_l0_raw']
    new_weights['0.rnns.1.module.bias_ih_l0']=old_weights['0.rnns.1.module.bias_ih_l0']
    new_weights['0.rnns.1.module.bias_hh_l0']=old_weights['0.rnns.1.module.bias_hh_l0']
    new_weights['0.rnns.2.weight_hh_l0_raw']=old_weights['0.rnns.2.module.weight_hh_l0_raw']
    new_weights['0.rnns.2.module.weight_ih_l0']=old_weights['0.rnns.2.module.weight_ih_l0']
    new_weights['0.rnns.2.module.weight_hh_l0']=old_weights['0.rnns.2.module.weight_hh_l0_raw']
    new_weights['0.rnns.2.module.bias_ih_l0']=old_weights['0.rnns.2.module.bias_ih_l0']
    new_weights['0.rnns.2.module.bias_hh_l0']=old_weights['0.rnns.2.module.bias_hh_l0']
    new_weights['1.decoder.weight'] = old_weights['1.decoder.weight']
    return new_weights


def pth_to_torch(old_weights: OrderedDict):
    return OrderedDict(old_weights['model'])



from torch.nn import Embedding, ModuleList

from typing import List, Tuple, Optional

import torch
from torch import Tensor
from fastai.text import SequentialRNN, AWD_LSTM, WeightDropout, EmbeddingDropout, RNNDropout, one_param, to_detach, \
    Module

TORCH_LONG_MIN_VAL = -2 ** 63


def to_test_mode(model: SequentialRNN) -> None:
    # Set batch size to 1
    model[0].bs = 1
    # Turn off dropout
    model.eval()
    # Reset hidden state
    model.reset()


def save_hidden_states(model: SequentialRNN) -> List[Tuple[torch.Tensor, torch.Tensor]]:
    return [(hl[0].clone(), hl[1].clone()) if isinstance(hl, (tuple, list)) else hl.clone() for hl in model[0].hidden]


def get_last_layer_activations(model: SequentialRNN, input: torch.FloatTensor) -> Optional[torch.FloatTensor]:
    tensor_rank = len(input.size())
    if tensor_rank != 2:
        if tensor_rank == 0:
            error_msg = 'Your input tensor is a scalar. If you want to use batch size = 1 and feed only one token to the model, pass input[None, None] to this method.'
        elif tensor_rank == 1:
            error_msg = 'You input tensor has rank 1. If you were intending to use batch size = 1, please pass input[None, :] to this method. If you however were intending to use multiple batches but feed only one element to the model pass input[:, None]'
        else:
            error_msg = f'Your tensor has rank {tensor_rank}.'

        raise ValueError(f'This method accepts tensors of rank 2. {error_msg}')

    if input.nelement() == 0:
        return None

    last_layer_activations, *_ = model(input)
    return last_layer_activations


class GRU(Module):

    initrange=0.1

    def __init__(self, vocab_sz:int, emb_sz:int, n_hid:int, n_layers:int, pad_token:int=1, hidden_p:float=0.2,
                 input_p:float=0.6, embed_p:float=0.1, weight_p:float=0.5, qrnn:bool=False, bidir:bool=False):
        self.bs,self.qrnn,self.emb_sz,self.n_hid,self.n_layers = 1,qrnn,emb_sz,n_hid,n_layers
        self.n_dir = 2 if bidir else 1
        self.encoder = Embedding(vocab_sz, emb_sz, padding_idx=pad_token)
        self.encoder_dp = EmbeddingDropout(self.encoder, embed_p)

        self.rnns = [torch.nn.GRU(emb_sz if l == 0 else n_hid, (n_hid if l != n_layers - 1 else emb_sz)//self.n_dir, 1,
                             batch_first=True, bidirectional=bidir) for l in range(n_layers)]
        self.rnns = [WeightDropout(rnn, weight_p) for rnn in self.rnns]
        self.rnns = ModuleList(self.rnns)
        self.encoder.weight.data.uniform_(-self.initrange, self.initrange)
        self.input_dp = RNNDropout(input_p)
        self.hidden_dps = ModuleList([RNNDropout(hidden_p) for l in range(n_layers)])

    def forward(self, input:Tensor, from_embeddings:bool=False)->Tuple[Tensor,Tensor]:
        if from_embeddings: bs,sl,es = input.size()
        else: bs,sl = input.size()
        if bs!=self.bs:
            self.bs=bs
            self.reset()
        raw_output = self.input_dp(input if from_embeddings else self.encoder_dp(input))
        new_hidden,raw_outputs,outputs = [],[],[]
        for l, (rnn,hid_dp) in enumerate(zip(self.rnns, self.hidden_dps)):
            raw_output, new_h = rnn(raw_output, self.hidden[l])
            new_hidden.append(new_h)
            raw_outputs.append(raw_output)
            if l != self.n_layers - 1: raw_output = hid_dp(raw_output)
            outputs.append(raw_output)
        self.hidden = to_detach(new_hidden, cpu=False)
        return raw_outputs, outputs

    def _one_hidden(self, l:int)->Tensor:
        "Return one hidden state."
        nh = (self.n_hid if l != self.n_layers - 1 else self.emb_sz) // self.n_dir
        return one_param(self).new(self.n_dir, self.bs, nh).zero_()

    def select_hidden(self, idxs):
        self.hidden = [h[:,idxs,:] for h in self.hidden]
        self.bs = len(idxs)

    def reset(self):
        "Reset the hidden states."
        [r.reset() for r in self.rnns if hasattr(r, 'reset')]
        self.hidden = [self._one_hidden(l) for l in range(self.n_layers)]


def add_gru_to_model_data():
    from fastai.text.learner import _model_meta
    gru_meta_data = {k:v for k, v in _model_meta[AWD_LSTM].items()}
    _model_meta[GRU] = gru_meta_data


import colorsys
from math import log

from langmodels.evaluation.metrics import entropy_to_probability

MIN_HUE = 0.0
MAX_HUE = 0.33

LIGHT = 0.4
SATURATION = 0.7

EPS = 0.05
NEUTRAL_RANK = 4
NEUTRAL_PROBABILITY = 0.25


def normalize_rank(rank: int) -> float:
    """
    >>> normalize_rank(1)
    1.0

    >>> normalize_rank(2)
    0.41421356237309515

    >>> normalize_rank(4)
    0.0

    >>> normalize_rank(10)
    -0.3675444679663241

    >>> normalize_rank(100)
    -0.8

    """
    assert rank > 0

    return rank ** log(0.5, NEUTRAL_RANK) * 2 - 1


def normalize_entropy(entropy: float) -> float:
    """
    >>> normalize_entropy(0.0)
    1.0

    >>> normalize_entropy(0.83)
    0.5000389892858181

    >>> normalize_entropy(2.0)
    0.0

    >>> normalize_entropy(100.0)
    -0.9999999999999982

    >>> normalize_entropy(1e+6)
    -1.0

    """
    assert entropy >= 0.0

    prob = entropy_to_probability(entropy)
    result = prob ** log(0.5, NEUTRAL_PROBABILITY) * 2 - 1

    assert -1.0 <= result <= 1.0

    return result


def get_hex_color(metric: float) -> str:
    """
    >>> get_hex_color(0.0)
    '000000'

    >>> get_hex_color(0.09)
    'a2ad1f'

    >>> get_hex_color(0.1)
    'a1ad1f'

    >>> get_hex_color(-0.1)
    'ad9e1f'

    >>> get_hex_color(1.0)
    '21ad1f'

    >>> get_hex_color(-1.0)
    'ad1f1f'
    """
    assert -1.0 <= metric <= 1.0

    if abs(metric) < EPS:
        return '000000'

    neutral = (MAX_HUE + MIN_HUE) / 2.0
    span = (MAX_HUE - MIN_HUE) / 2.0

    rgb_normalized = colorsys.hls_to_rgb(neutral + span * metric, LIGHT, SATURATION)
    rgb = tuple(round(255 * c) for c in rgb_normalized)

    return ''.join(f'{i:02x}' for i in rgb)


def improvement_metric(before: float, after: float) -> float:
    assert -1.0 <= before <= 1.0
    assert -1.0 <= after <= 1.0

    return (after - before) / 2.0



import logging
import os
from collections import OrderedDict
from threading import Lock

from dataclasses import dataclass, asdict

from dataprep.subtokens import is_terminal_subtoken

from dataprep.infrastructure.dataset import normalize_extension_string

from dataprep.parse.model.metadata import PreprocessingMetadata, check_metadata_validity
from math import exp
from typing import List, Dict, Any, Tuple, Optional, Union

import torch
from dataprep.api.corpus import PreprocessedCorpus
from dataprep.parse.model.placeholders import placeholders
from fastai.text import SequentialRNN, get_language_model, F, Vocab, awd_lstm_lm_config, convert_weights
from fastai.text.models.transformer import init_transformer
from torch import cuda

from langmodels.beamsearch import beam_search
from langmodels.lmconfig.datamodel import Corpus, LstmArch, TransformerArch, LMTrainingConfig, GruArch, \
    LMTrainingMetrics
from langmodels.lmconfig.serialization import load_config_from_file, read_value_from_file
from langmodels.migration import pth_to_torch
from langmodels.nn import to_test_mode, get_last_layer_activations, TORCH_LONG_MIN_VAL
from langmodels.util import to_binary_entropy
from langmodels.cuda_util import get_device, get_map_location

logger = logging.getLogger(__name__)


PAD_TOKEN_INDEX = 1


def create_custom_lstm_or_gru_config(arch: Union[GruArch, LstmArch]):
    config = awd_lstm_lm_config
    config['emb_sz'] = arch.emb_sz
    config['n_hid'] = arch.n_hid
    config['n_layers'] = arch.n_layers
    config['pad_token'] = PAD_TOKEN_INDEX
    if isinstance(arch, LstmArch):
        config['qrnn'] = arch.qrnn
    config['bidir'] = arch.bidir
    config['output_p'] = arch.drop.out
    config['hidden_p'] = arch.drop.outh
    config['input_p'] = arch.drop.outi
    config['embed_p'] = arch.drop.oute
    config['weight_p'] = arch.drop.w
    config['tie_weights'] = arch.tie_weights
    config['out_bias'] = arch.out_bias
    return config


def create_custom_transformer_config(arch: TransformerArch) -> Dict[str, Any]:
    d = {'init': init_transformer,
         'ctx_len': arch.ctx_len,
         'n_layers': arch.n_layers,
         'n_heads': arch.n_heads,
         'd_model': arch.d_model,
         'd_head': arch.d_head,
         'd_inner': arch.d_inner,
         'resid_p': arch.drop.resid,
         'attn_p': arch.drop.attn,
         'ff_p': arch.drop.ff,
         'embed_p': arch.drop.embed,
         'output_p': arch.drop.output,
         'bias': arch.bias,
         'scale': arch.scale,
         'act': arch.act,
         'double_drop': arch.double_drop,
         'tie_weights': arch.tie_weights,
         'out_bias': arch.out_bias,
         'mask': arch.mask,
    }
    return d


def create_custom_config(lm_training_config: LMTrainingConfig):
    arch = lm_training_config.arch
    if isinstance(arch, LstmArch) or isinstance(arch, GruArch):
        return create_custom_lstm_or_gru_config(arch)
    elif isinstance(arch, TransformerArch):
        return create_custom_transformer_config(arch)
    else:
        raise ValueError(f"Unknown architecture: {arch}")


def _create_term_vocab(vocab: Vocab) -> Tuple[Vocab, int]:
    terminal_token_indices = {i for i, k in enumerate(vocab.itos) if is_terminal_subtoken(k)}
    term_vocab_list = [vocab.itos[i] for i in terminal_token_indices]
    non_term_vocab_list = [vocab.itos[i] for i in range(len(vocab.itos)) if i not in terminal_token_indices]
    term_vocab = Vocab(term_vocab_list + non_term_vocab_list)
    return term_vocab, len(term_vocab_list)


def _check_is_term_vocab(vocab: Vocab, first_non_term: int) -> None:
    for token in vocab.itos[:first_non_term]:
        if not is_terminal_subtoken(token):
            raise ValueError(f"First non-terminal token in vocab has index {first_non_term}, "
                             f"however there's a non-terminal token {token} "
                             f"that has index {vocab.itos.index(token)}")
    for token in vocab.itos[first_non_term:]:
        if is_terminal_subtoken(token):
            raise ValueError(f"Starting from index {first_non_term} there should bb only non-term "
                             f"tokens in the vocab, however there's a terminal token {token} "
                             f"that has index {vocab.itos.index(token)}")


DEFAULT_BEAM_SIZE = 500


def to_full_token_string(vocab: Vocab, subtokens_num: torch.LongTensor, include_debug_tokens=False) -> str:
    try:
        subtokens_num = subtokens_num[:subtokens_num.tolist().index(TORCH_LONG_MIN_VAL)]
    except ValueError: pass

    sep = '|' if include_debug_tokens else ''
    textified = vocab.textify(subtokens_num, sep=sep)
    cwe = placeholders['compound_word_end']
    if textified in placeholders.values():
        return textified

    if not is_terminal_subtoken(textified):
        raise ValueError(f'{textified} ({subtokens_num}) is not a full token')
    return textified if include_debug_tokens else textified[:-len(cwe)]


@dataclass
class ModelDescription(object):
    id: str
    bpe_merges: str
    layers_config: str
    arch: str
    bin_entropy: float
    training_time_minutes_per_epoch: int
    n_epochs: int
    best_epoch: int
    tags: List[str]

    def is_tagged_by(self, tag: str) -> bool:
        return tag in self.tags

    @staticmethod
    def get_attribute_list() -> List[str]:
        return [k for k in ModelDescription.__annotations__.keys()]

    def get_value_list(self) -> List[str]:
        return list(map(lambda a: self.__getattribute__(a), ModelDescription.get_attribute_list()))


lock = Lock()


class TrainedModel(object):
    STARTING_TOKEN = placeholders['ect']

    def __init__(self, path: str, force_use_cpu: bool = False, load_only_description: bool = False):
        self.force_use_cpu = force_use_cpu
        self.id = os.path.basename(path)
        path_to_config_file = os.path.join(path, 'config')
        path_to_metrics_file = os.path.join(path, 'metrics')
        path_to_tags_file = os.path.join(path, 'tags')
        self.metrics = None
        self.config = None
        self.tags = []
        try:
            self.config: LMTrainingConfig = load_config_from_file(path_to_config_file)
        except FileNotFoundError:
            logger.warning(f'Config file not found: {path_to_config_file}')
        try:
            self.metrics: LMTrainingMetrics = load_config_from_file(os.path.join(path, 'metrics'))
        except FileNotFoundError:
            logger.warning(f'File with metrics not found: {path_to_metrics_file}')
        if os.path.exists(path_to_tags_file):
            value = read_value_from_file(path_to_tags_file, value_type=str)
            if value != '':
                self.tags = value.split(',')
        self._prep_function = self.config.prep_function

        self.load_only_description = load_only_description
        if not load_only_description:
            # we might want to load only description without loading actual weights when we want
            # to save time when loading multiple models to choose one of them to work with

            self.original_vocab = Vocab.load(os.path.join(path, 'vocab'))
            term_vocab, self.first_nonterm_token = _create_term_vocab(self.original_vocab)
            self.model, self.vocab = self._load_model(path, term_vocab)
            to_test_mode(self.model)

            # last_predicted_token_tensor is a rank-2 tensor!
            self.last_predicted_token_tensor = torch.tensor([self.vocab.numericalize([self.STARTING_TOKEN])],
                                                            device=get_device(self.force_use_cpu))
            self.beam_size = DEFAULT_BEAM_SIZE

    def _load_model(self, path: str, custom_vocab: Optional[Vocab] = None) -> Tuple[SequentialRNN, Vocab]:
        path_to_model = os.path.join(path, 'best.pth')
        logger.debug(f"Loading model from: {path_to_model} ...")

        vocab = custom_vocab if custom_vocab else self.original_vocab
        model = get_language_model(self.config.get_arch_class(), len(vocab.itos), create_custom_config(self.config))
        map_location = get_map_location(self.force_use_cpu)
        if cuda.is_available():
            model.cuda()
        state_dict = torch.load(path_to_model, map_location=map_location)

        # a more simple solution is to use fastai's load_learner,
        # however it doesn't seem to work out of the box with customizations we've done

        # update: it appeared that it's quite simple to load weights. So maybe not worth it loading a learner?
        weights = OrderedDict(state_dict['model'] if ('model' in state_dict) else state_dict)
        if custom_vocab:
            weights = convert_weights(weights, self.original_vocab.stoi, custom_vocab.itos)
        model.load_state_dict(weights, strict=True)

        return model, vocab

    def prep_corpus(self, corpus: Corpus, **kwargs) -> PreprocessedCorpus:
        return self._prep_function.apply(corpus, **kwargs)

    def prep_text(self, text, **kwargs) -> Union[Tuple[List[str], PreprocessingMetadata], List[str]]:
        import dataprep.api.text as text_api
        text_callable = getattr(text_api, self._prep_function.callable.__name__)
        prep_text, metadata = text_callable(text, *self._prep_function.params, **asdict(self._prep_function.options), **kwargs)
        check_metadata_validity(prep_text, metadata)
        return (prep_text, metadata) if 'return_metadata' in kwargs and kwargs['return_metadata'] else prep_text

    def _check_model_loaded(self, only_description=False):
        if not only_description and self.load_only_description:
            raise RuntimeError("Operation not supported. Only model's description is loaded. "
                               "Prease reload the model with param load_only_description set to False.")

    def feed_text(self, text: str) -> None:
        self._check_model_loaded()

        prep_text, metadata = self.prep_text(text, return_metadata=True, force_reinit_bpe_data=False)
        context_tensor = torch.tensor([self.vocab.numericalize(prep_text)], device=get_device(self.force_use_cpu))
        with lock:
            _ = get_last_layer_activations(self.model, context_tensor[:,:-1])
            self.last_predicted_token_tensor = context_tensor[:,-1:]

    def get_entropies_for_prep_text(self, prep_text: List[str]) -> List[float]:
        """
        changes hidden states of the model!!
        """
        with lock:
            self._check_model_loaded()

            if not prep_text:
                return []

            loss_list = []
            max_subtokens_per_chunk = 1000
            # if the line is too big, we break it down to chunks to fit it into gpu memory
            # big chunks require more memory, small chunks require more time
            n_chunks = (len(prep_text)-1) // max_subtokens_per_chunk + 1
            for i in range(n_chunks):
                pt = prep_text[i*max_subtokens_per_chunk:(i+1)*max_subtokens_per_chunk]
                numericalized_prep_text = torch.tensor([self.vocab.numericalize(pt)],
                                                       device=get_device(self.force_use_cpu))

                last_layer = get_last_layer_activations(self.model, torch.cat([self.last_predicted_token_tensor, numericalized_prep_text[:, :-1]], dim=1))
                loss = F.cross_entropy(last_layer.view(-1, last_layer.shape[-1]),
                                       numericalized_prep_text.view(-1),
                                       reduction='none')
                binary_loss = to_binary_entropy(loss)
                loss_list.extend(binary_loss.tolist())
                self.last_predicted_token_tensor = numericalized_prep_text[:, -1:]
            return loss_list

    def reset(self) -> None:
        with lock:
            self._check_model_loaded()

            self.model.reset()
            self.last_predicted_token_tensor = torch.tensor([self.vocab.numericalize([self.STARTING_TOKEN])],
                                                        device=get_device(self.force_use_cpu))

    def predict_next_full_token(self, n_suggestions: int, include_debug_tokens: bool = False) -> List[Tuple[str, float]]:
        with lock:
            self._check_model_loaded()

            subtokens, scores = beam_search(self.model, self.last_predicted_token_tensor[0], self.first_nonterm_token, n_suggestions, self.beam_size)
            return [(to_full_token_string(self.vocab, st, include_debug_tokens=include_debug_tokens), 1 / exp(score.item())) for st, score in zip(subtokens, scores)]

    def _format_layers_config(self) -> str:
        if isinstance(self.config.arch, TransformerArch):
            return "transformer"  # TODO add proper layer description
        n_layers = self.config.arch.n_layers
        emb_size = self.config.arch.emb_sz
        n_hid = self.config.arch.n_hid
        return f'{emb_size}/{n_layers}/{n_hid}'

    def get_model_description(self) -> ModelDescription:
        return ModelDescription(id=self.id,
                                bpe_merges=self.config.prep_function.params[0],
                                layers_config=self._format_layers_config(),
                                arch=str(self.config.get_arch_class().__name__),
                                bin_entropy=self.metrics.bin_entropy if self.metrics else 1e+6,
                                training_time_minutes_per_epoch=self.metrics.training_time_minutes_per_epoch
                                if self.metrics else 0,
                                n_epochs=self.metrics.n_epochs if self.metrics else 0,
                                best_epoch=self.metrics.best_epoch if self.metrics else -1,
                                tags=self.tags)

    def check_inference_possible_for_file_type(self, extension):
        if extension not in normalize_extension_string(self.config.corpus.extensions):
            raise ValueError(f'The model was not trained on .{extension} files. Cannot do inference.')

    def __str__(self) -> str:
        return str(self.get_model_description())



from typing import Tuple, List, Dict

import torch
from fastai.text import SequentialRNN
from torch.nn.functional import log_softmax

from langmodels.nn import get_last_layer_activations, save_hidden_states, TORCH_LONG_MIN_VAL
from langmodels.cuda_util import get_device

DEVICE = get_device()


def get_topk_predictions(model: SequentialRNN, context: torch.FloatTensor, top_k: int) -> Tuple[
    torch.FloatTensor, torch.LongTensor]:
    last_token_activations = get_last_layer_activations(model, context)
    predictions = log_softmax(last_token_activations[:, -1], dim=-1)  # TODO log_softmax not really needed
    return predictions.topk(top_k, dim=-1)


def topk_are_full_tokens(full_tokens_bitmap: torch.Tensor, top_k: int) -> bool:
    return full_tokens_bitmap[:top_k].sum() == top_k


Tensors = Dict[str, torch.Tensor]


def get_new_candidates(model: SequentialRNN, context: torch.LongTensor, n_predictions: int,
                       pending_candidates: Tensors) -> Tensors:
    assert len(context.size()) == 2

    loss, num_tokens = get_topk_predictions(model, context, n_predictions)
    batch_size = context.size(0)

    new_candidates = {'scores': (-loss + pending_candidates['scores'][:, None]).view(-1),
                      'num_tokens': num_tokens.view(-1),
                      'hidden_indices': torch.arange(0, batch_size, device=DEVICE)[:, None] \
                          .expand(batch_size, n_predictions).contiguous().view(-1),
                      'tokens': pending_candidates['tokens'].repeat(n_predictions, 1)}
    return new_candidates


def update_ready_candidates(best_subtokens: torch.LongTensor, best_scores: torch.FloatTensor) -> Tensors:
    ready_candidates = {'tokens': best_subtokens, 'scores': best_scores}
    n_full_tokens = best_scores.size(0)
    ready_candidates['num_tokens'] = torch.full((n_full_tokens,), fill_value=TORCH_LONG_MIN_VAL, dtype=torch.long, device=DEVICE)
    ready_candidates['hidden_indices'] = torch.full((n_full_tokens,), fill_value=TORCH_LONG_MIN_VAL, dtype=torch.long, device=DEVICE)
    return ready_candidates


def beam_search(model: SequentialRNN, context: torch.LongTensor, first_nonterm_token: int, top_k: int,
                beam_size: int) -> List[Tuple]:
    if len(context.size()) != 1:
        raise ValueError("The rank of context tensor should be one. Beam search is only possible with batch size = 1.")

    context = context.unsqueeze(dim=0)

    checkpoint = save_hidden_states(model)

    full_tokens_bitmap_sorted = torch.zeros(1, device=DEVICE)
    pending_candidates_sorted = {'tokens': torch.empty(context.size(0), 0, device=DEVICE).long(),
                                 'scores': torch.zeros(1, device=DEVICE).float()}
    ready_candidates = update_ready_candidates(
        best_subtokens=torch.empty((0, 0), dtype=torch.long, device=DEVICE),
        best_scores=torch.empty((0), dtype=torch.float, device=DEVICE)
    )
    with torch.no_grad():
        while not topk_are_full_tokens(full_tokens_bitmap_sorted, top_k):
            new_candidates = get_new_candidates(model, context, beam_size, pending_candidates_sorted)

            all_candidates = {k: torch.cat([ready_candidates[k], new_candidates[k]], dim=0) for k in
                              ready_candidates.keys()}

            all_candidates['tokens'] = torch.cat([all_candidates['tokens'], all_candidates['num_tokens'][:, None]],
                                                 dim=-1)

            idx_in_score_asc_order = all_candidates['scores'].argsort()[:beam_size]
            best_candidates_sorted = {k: all_candidates[k][idx_in_score_asc_order] for k in all_candidates.keys()}

            full_tokens_bitmap_sorted = best_candidates_sorted['num_tokens'] < first_nonterm_token
            ready_candidate_idxs = full_tokens_bitmap_sorted.nonzero().squeeze(dim=1)
            pending_candidate_idxs = (full_tokens_bitmap_sorted == 0).nonzero().squeeze(dim=1)

            ready_candidates = update_ready_candidates(
                best_candidates_sorted['tokens'][ready_candidate_idxs],
                best_candidates_sorted['scores'][ready_candidate_idxs]
            )

            pending_candidates_sorted = {k: best_candidates_sorted[k][pending_candidate_idxs] for k in
                                         best_candidates_sorted.keys()}

            model[0].select_hidden(pending_candidates_sorted['hidden_indices'])
            context = pending_candidates_sorted['num_tokens'].unsqueeze(dim=1)

    model[0].hidden = checkpoint
    model[0].bs = checkpoint[0][0].size(0)

    return ready_candidates['tokens'][:top_k], ready_candidates['scores'][:top_k]



from unittest import mock

from pytest_mock import MockFixture

from dataprep.parse.model.metadata import PreprocessingMetadata
from unittest.mock import MagicMock, Mock

from langmodels.evaluation import evaluate_model_on_string
from langmodels.evaluation.filtering import TokenTypes
from langmodels.evaluation.metrics import EvaluationResult, EvaluationScenario, Evaluation
from langmodels.model import TrainedModel


def test_evaluate_model_on_string_empty():
    trained_model_mock = MagicMock(spec=TrainedModel)
    trained_model_mock.prep_text.return_value = ([], PreprocessingMetadata())

    expected = [Evaluation(
        '', [], PreprocessingMetadata(),
        {EvaluationScenario('full_token_entropy', TokenTypes.ALL): EvaluationResult([], 0., 0)}
    )]
    actual = evaluate_model_on_string(trained_model_mock, '')

    assert actual == expected


def test_evaluate_on_string_default_args(mocker: MockFixture):
    # given
    text = 'MyClass'
    prep_line = ['My', 'Class</t>']
    metadata = PreprocessingMetadata(word_boundaries=[0, 2])
    scenarios = {EvaluationScenario('full_token_entropy', TokenTypes.ALL): EvaluationResult([1.0, 2.0], 3.0, 1)}

    trained_model_mock = Mock(spec=TrainedModel)
    trained_model_mock.prep_text.return_value = (prep_line, metadata)

    mocked_metric = Mock(spec=callable, return_value=scenarios)
    mocker.patch('langmodels.evaluation.evaluation.DEFAULT_METRIC', new=mocked_metric)

    # when
    actual = evaluate_model_on_string(trained_model_mock, text)

    # then
    trained_model_mock.prep_text.assert_has_calls([mock.call('MyClass', extension='java',
                                                             force_reinit_bpe_data=False,
                                                             return_metadata=True, append_eof=False)])
    mocked_metric.assert_called_with(trained_model_mock, prep_line, metadata, None)
    assert actual == [Evaluation(text, prep_line, metadata, scenarios)]


def test_evaluate_on_string_default_args_not_result_per_line(mocker: MockFixture):
    # given
    text = 'MyClass\n{'
    prep_line = ['My', 'Class</t>']
    metadata = Mock(spec=PreprocessingMetadata)
    scenarios = {EvaluationScenario('full_token_entropy', TokenTypes.ALL): Mock(spec=EvaluationResult)}

    trained_model_mock = Mock(spec=TrainedModel)
    trained_model_mock.prep_text.return_value = (prep_line, metadata)

    mocked_metric = Mock(spec=callable, return_value=scenarios)
    mocker.patch('langmodels.evaluation.evaluation.DEFAULT_METRIC', new=mocked_metric)

    # when
    actual = evaluate_model_on_string(trained_model_mock, text, result_per_line=False)

    # then
    trained_model_mock.prep_text.assert_has_calls([mock.call('MyClass\n{', extension='java',
                                                             force_reinit_bpe_data=False,
                                                             return_metadata=True, append_eof=False)])
    mocked_metric.assert_called_with(trained_model_mock, prep_line, metadata, None)
    assert actual == [Evaluation(text, prep_line, metadata, scenarios)]


def test_evaluate_on_string_non_default_token_types_and_metrics_multiline(mocker: MockFixture):
    # given
    text = 'MyClass\n{'
    prep_lines = [['My', 'Class</t>'], ['{']]
    metadata_list = [Mock(spec=PreprocessingMetadata) for i in range(len(prep_lines))]
    metrics = {'full_token_entropy', 'mrr'}
    token_types_list = {TokenTypes.ALL, TokenTypes.ALL_BUT_COMMENTS}
    scenarios = [[{EvaluationScenario(metric, token_types): Mock(spec=EvaluationResult)
                   for token_types in token_types_list} for i in range(len(prep_lines))] for metric in metrics]
    mocked_metrics = [Mock(spec=callable, side_effect=scenario) for scenario in scenarios]

    trained_model_mock = Mock(spec=TrainedModel)
    trained_model_mock.prep_text.side_effect = list(zip(prep_lines, metadata_list))

    mocked_metrics_from_strings = Mock(spec=callable, return_value=mocked_metrics)
    mocker.patch('langmodels.evaluation.evaluation.metrics_from_strings', new=mocked_metrics_from_strings)

    # when
    actual = evaluate_model_on_string(trained_model_mock, text,
                                      token_types=token_types_list,
                                      metrics=metrics,
                                      result_per_line=True, append_eof=True)

    # then
    trained_model_mock.prep_text.assert_has_calls([mock.call('MyClass', extension='java',
                                                             force_reinit_bpe_data=False,
                                                             return_metadata=True, append_eof=False)])
    trained_model_mock.prep_text.assert_has_calls([mock.call('{', extension='java',
                                                             force_reinit_bpe_data=False,
                                                             return_metadata=True, append_eof=True)])

    for mocked_metric in mocked_metrics:
        for prep_line, metadata in zip(prep_lines, metadata_list):
            mocked_metric.assert_has_calls([mock.call(trained_model_mock, prep_line, metadata, token_types_list)])

    assert actual == [Evaluation('MyClass', prep_lines[0], metadata_list[0],
                                 {k: v for ss in [scenarios[0][0], scenarios[1][0]] for k, v in ss.items()}),
                      Evaluation('{', prep_lines[1], metadata_list[1],
                                 {k: v for ss in [scenarios[0][1], scenarios[1][1]] for k, v in ss.items()})]



import logging
import os
from pathlib import Path

from tqdm import tqdm
from typing import List, Tuple, Callable, Optional, Union, Dict, Set

from langmodels.file_util import get_all_files, read_file_contents, get_file_extension
from langmodels.evaluation.metrics import bin_entropy, mrr, get_metric_aggregator_by_name, Metric, \
    get_metric_func_by_name, TokenTypes, EvaluationScenario, Evaluation
from langmodels.model import TrainedModel

logger = logging.getLogger(__name__)


def zip_subwords(subwords: Tuple[List[str], List[str]],
                 entropies: Tuple[List[float], List[float]]
                 ) -> Tuple[List[str], List[Tuple[float, float]]]:
    """
    >>> zip_subwords((['a'], ['a']), ([1.0], [2.0]))
    (['a'], [(1.0, 2.0)])

    >>> zip_subwords((['a', 'b'], ['a', 'b']), ([1.0, 2.0], [3.0, 4.0]))
    (['a', 'b'], [(1.0, 3.0), (2.0, 4.0)])

    >>> zip_subwords((['a', 'b'], ['ab']), ([1.0, 2.0], [7.0]))
    (['ab'], [(3.0, 7.0)])

    >>> zip_subwords((['abcdef'], ['ab', 'cd', 'ef']), ([1.0], [7.0, 4.0, 5.0]))
    (['abcdef'], [(1.0, 16.0)])

    >>> zip_subwords((['a', 'bc', 'd'], ['ab', 'c', 'd']), ([1.0, 2.0, 7.0], [4.0, 5.0, 78.0]))
    (['abc', 'd'], [(3.0, 9.0), (7.0, 78.0)])

    >>> zip_subwords((['a'], ['ab']), ([1.0], [2.0]))
    Traceback (most recent call last):
    ...
    AssertionError
    """
    res: Tuple[List[str], List[Tuple[float, float]]] = ([], [])
    indices = [0, 0]
    next_subwords = tuple([s[i] for s, i in zip(subwords, indices)])
    entropy_sums = [.0, .0]
    prefix = ""
    while True:
        if next_subwords[0] == next_subwords[1]:
            res[0].append(prefix + next_subwords[0])
            res[1].append(tuple(
                e + es[i] for e, es, i in zip(entropy_sums, entropies, indices)
            ))
            entropy_sums = [.0, .0]
            indices = tuple(i+1 for i in indices)
            if indices[1] == len(subwords[1]):
                break
            next_subwords = tuple(s[i] for s, i in zip(subwords, indices))
            prefix = ""
        elif not next_subwords[0].startswith(next_subwords[1]) and \
                not next_subwords[1].startswith(next_subwords[1]):
            raise ValueError(f"Different words passed: {prefix}{next_subwords[1]} != {prefix}{next_subwords[0]}")
        else:
            if next_subwords[0].startswith(next_subwords[1]):
                longer = 0
                shorter = 1
            else:
                longer = 1
                shorter = 0

            entropy_sums[shorter] += entropies[shorter][indices[shorter]]
            indices[shorter] += 1
            prefix += next_subwords[shorter]
            if indices[shorter] == len(subwords[shorter]):
                break
            first_common_chars = len(next_subwords[shorter])
            next_subwords = (
                next_subwords[longer][first_common_chars:],
                subwords[shorter][indices[shorter]]
            )
            if longer == 1:
                next_subwords = next_subwords[1], next_subwords[0]

    for i, s in zip(indices, subwords):
        assert i == len(s)

    return res


def metrics_from_strings(_metrics: Union[Set[str], Set[Callable], None]) -> Optional[Set[Callable]]:
    if _metrics is None:
        return None

    return set(map(lambda m: m if isinstance(m, Callable) else get_metric_func_by_name(m), _metrics))


DEFAULT_METRIC = bin_entropy


def evaluate_model_on_string(model: TrainedModel, text: str, extension='java',
                             token_types: Optional[Set[TokenTypes]] = None,
                             metrics: Optional[Union[Set[Metric], Set[str]]] = None,
                             result_per_line=True, append_eof: bool = False) -> List[Evaluation]:
    metrics = metrics_from_strings(metrics) or {DEFAULT_METRIC}

    show_progress = result_per_line and mrr in metrics

    text_lines = text.split('\n') if result_per_line else [text]
    model_evaluation: List[Evaluation] = []
    line_iterable = tqdm(text_lines) if show_progress else text_lines
    for i, line in enumerate(line_iterable):
        prep_line, metadata = model.prep_text(line, return_metadata=True,
                                              force_reinit_bpe_data=False,
                                              extension=extension,
                                              append_eof=append_eof and i == len(line_iterable) - 1)
        scenarios = {k: v for metric in metrics for k, v in metric(model, prep_line, metadata, token_types).items()}
        line_evaluation = Evaluation(text=line, prep_text=prep_line, prep_metadata=metadata, scenarios=scenarios)
        model_evaluation.append(line_evaluation)
    return model_evaluation


def evaluate_model_on_file(model: TrainedModel, file: Path,
                           token_types: Optional[Set[TokenTypes]] = None,
                           metrics: Optional[Union[Set[Metric], Set[str]]] = None,
                           result_per_line: bool = True) -> Union[List[Evaluation], Evaluation]:
    suffix: str = file.suffix[1:]
    model.check_inference_possible_for_file_type(suffix)
    text = read_file_contents(file)
    result = evaluate_model_on_string(model, text, suffix, token_types, metrics,
                                      result_per_line=result_per_line, append_eof=True)
    return result if result_per_line else result[0]


def _format_postfix(current_metrics: Dict[EvaluationScenario, Tuple[float, int]]) -> Dict[str, str]:
    if current_metrics:
        postfix = {str(eval_scenario): f'{value:.2f} (n={n_samples})'
                   for eval_scenario, (value, n_samples) in current_metrics.items()}
    else:
        postfix = {}
    return postfix


def evaluate_model_on_project_set(model: TrainedModel, path: str,
                                  token_types: Optional[Set[TokenTypes]] = None,
                                  metrics: Optional[List[Metric]] = None) \
        -> Dict[str, Dict[EvaluationScenario, Tuple[float, int]]]:
    result: Dict[str, Dict[EvaluationScenario, Tuple[float, int]]] = {}
    root, dirs, files = next(os.walk(path, followlinks=True))
    for directory in dirs:
        logger.info(f'Evaluating {directory} ...')
        result[directory] = evaluate_model_on_path(model, os.path.join(root, directory), token_types, metrics)
    return result


def evaluate_model_on_path(model: TrainedModel, path: str, token_types: Optional[Set[TokenTypes]] = None,
                           metrics: Optional[Union[Set[Metric], Set[str]]] = None) \
        -> Dict[EvaluationScenario, Tuple[float, int]]:
    token_types = token_types or {TokenTypes.ALL}

    logger.info("Counting total file number ...")
    all_files = [f for f in get_all_files(path)]

    cumulative_metrics: Dict[EvaluationScenario, Tuple[float, int]] = {}
    t = tqdm(all_files)
    for file in t:
        postfix = _format_postfix(cumulative_metrics)
        t.set_postfix(postfix)
        evaluation: Evaluation = evaluate_model_on_file(model, file, token_types, metrics, result_per_line=False)

        current_file_metrics = {scenario: (eval_result.average, eval_result.n_samples)
                                for scenario, eval_result in evaluation.scenarios.items()}

        if cumulative_metrics:
            for scenario, cumulative_eval_result in cumulative_metrics.items():
                metric_agg = get_metric_aggregator_by_name(scenario.metric_name)
                cumulative_metrics[scenario] = metric_agg(*zip(cumulative_eval_result, current_file_metrics[scenario]))
        else:
            cumulative_metrics = current_file_metrics
    if not cumulative_metrics:
        raise Exception(f"No files to evaluate are found in {path}")

    return cumulative_metrics



import logging
import os

from pathlib import Path
from typing import Generator, Optional

logger = logging.getLogger(__name__)


def check_path_writable(path: str) -> None:
    if not os.path.exists(path):
        os.makedirs(path)
    elif not os.access(path, os.W_OK | os.X_OK):
        raise FileNotFoundError(path)


def check_path_exists(path: str) -> None:
    if not os.path.exists(path):
        raise FileNotFoundError(path)


def get_all_files(path: str, extension: Optional[str] = 'java') -> Generator[Path, None, None]:
    if os.path.isfile(path):
        yield Path(path)
    else:
        for root, dirs, files in os.walk(path, followlinks=True):
            for file in files:
                if not extension or file.endswith(f'.{extension}'):
                    yield Path(os.path.join(root, file))


def read_file_with_encoding(file: Path, encoding: str) -> str:
    with file.open('r', encoding=encoding) as f:
        return f.read()


def read_file_contents(file: Path) -> str:
    try:
        return read_file_with_encoding(file, 'utf-8')
    except UnicodeDecodeError:
        try:
            return read_file_with_encoding(file, 'ISO-8859-1')
        except UnicodeDecodeError:
            logger.error(f"Unicode decode error in file: {file}")


def get_file_total_lines(file):
    with open(file, 'r') as f:
        return len([i for i in f])


def get_file_extension(file: str) -> str:
    """
    >>> get_file_extension('new_file.java')
    'java'
    """
    return os.path.splitext(file)[1][1:]


import importlib
import importlib.util
# comet.ml must be imported before everything else
module = importlib.import_module('comet_ml')

import jsons

from langmodels.lmconfig.patch import patch_config
from langmodels.lmconfig.serialization import load_config_from_file

from typing import Dict, Optional, Any

from langmodels import app_name

import docopt_subcommands as dsc

from langmodels.training.training import train
from langmodels.cuda_util import CudaNotAvailable
from langmodels.lmconfig.datamodel import DeviceOptions, LMTrainingConfig
from langmodels import __version__


def get_option(args: Dict, option: str) -> Optional[Any]:
    return args[option] if option in args else None


def is_option_true(args: Dict, option: str) -> bool:
    return bool(get_option(args, option))


@dsc.command()
def train_handler(args):
    """usage: {program} train [--config <config> | --patch <patch>] [--fallback-to-cpu] [--tune] [--disable-comet] [--device=<device>]

    Trains a language model according to the given config.

    Options:
      -C, --fallback-to-cpu                        Fallback to cpu if gpu with CUDA-support is not available
      -x, --disable-comet                          Do not log experiment to comet.ml
      -t, --tune                                   Training will be done only on a few batches
                                                    (can be used for model params such as batch size to make sure
                                                    the model fits into memory)
      -d <device>, --device=<device>               Device id to use
      -c, --config=<config>                        Path to the json with config to be used to train the model
      -p, --patch=<patch>                          'Patch' to apply to the default lm training config e.g

    """
    handle_train(args)


def parse_patch(patch_string: str) -> Dict[str, str]:
    return {l[0]: l[1] for l in [param.split(':') for param in patch_string.split(',')]}


def handle_train(args) -> None:
    fallback_to_cpu = is_option_true(args, '--fallback-to-cpu')
    tune = is_option_true(args, '--tune')
    comet = not is_option_true(args, '--disable-comet')
    device = get_option(args, '--device')
    device = int(device) if device else 0
    path_to_config = get_option(args, '--config')
    try:
        lm_training_config = load_config_from_file(path_to_config) if path_to_config else LMTrainingConfig()
    except jsons.exceptions.DecodeError:
        raise ValueError(f"Could not deserialize a valid config from {path_to_config}")

    patch = get_option(args, '--patch')
    lm_training_config = patch_config(lm_training_config, parse_patch(patch)) if patch else lm_training_config

    device_options = DeviceOptions(fallback_to_cpu=fallback_to_cpu, non_default_device_to_use=device)

    try:
        train(training_config=lm_training_config, device_options=device_options, tune=tune, comet=comet)
    except CudaNotAvailable:
        print('Gpu with CUDA-support is not available on this machine. '
              'Use --fallback-to-cpu  switch if you want to train on gpu')
        exit(4)


def run(args):
    dsc.main(app_name, f'{app_name} {__version__}', argv=args, exit_at_end=False)



import logging
from typing import Any, Dict

from fastai.basic_train import Learner
from fastai.callbacks import TrackerCallback

logger = logging.getLogger(__name__)


class ReduceLRCallback(TrackerCallback):
    def __init__(self, learn: Learner, mult_coeff: float, max_times_lr_decrease: int, monitor: str = 'valid_loss',
                 mode: str = 'auto'):
        super().__init__(learn, monitor, mode)
        self.mult_coeff = mult_coeff
        self.lr_decreased_times = 0
        self.max_lr_decrease_times = max_times_lr_decrease

    def on_epoch_end(self, epoch: int, **kwargs: Any) -> Dict:
        current = self.get_monitor_value()
        if current is not None and self.operator(current, self.best):
            self.best = current
            self.lr_decreased_times = 0
        else:
            self.learn.opt.lr *= self.mult_coeff
            logger.info(
                f"Cross entropy loss increased, "
                f"reducing learning rate to {self.learn.opt.lr} for the next epoch ...")
            self.lr_decreased_times += 1
            if self.lr_decreased_times > self.max_lr_decrease_times:
                return {"stop_training": True}



import torch
from torch import Tensor


def _find_sorted_array_position(tensor: Tensor, values_tensor: Tensor) -> Tensor:
    dim0, dim1, dim2 = tensor.shape
    expanded_values_tensor = values_tensor.resize_((dim0, dim1, 1)).expand(dim0, dim1, dim2)
    position_of_value = torch.sum((tensor > expanded_values_tensor).long(), -1)
    return position_of_value.add(1)


def mrr(preds: Tensor, targs: Tensor) -> Tensor:
    """
    >>> preds = torch.tensor([[[0.2, 0.55, 0.25], [0.005, 0.005, 0.99]]])
    >>> targs = torch.tensor([[1, 2]])
    >>> mrr(preds, targs)
    tensor(1.)

    >>> preds = torch.tensor([[[0.2, 0.55, 0.25], [0.006, 0.004, 0.99]]])
    >>> targs = torch.tensor([[0, 0]])
    >>> mrr(preds, targs)
    tensor(0.4167)
    """
    pred_values = preds.gather(-1, targs.unsqueeze(-1))
    guessed_positions = _find_sorted_array_position(preds, pred_values).float()
    reciprocal = torch.reciprocal(guessed_positions)
    return torch.mean(reciprocal)


def contains_no_value(tensor: Tensor, value: int) -> bool:
    """
    >>> t = torch.full((100,100,), 2)
    >>> contains_no_value(t, 0)
    True

    >>> t = torch.full((100,100,), 2)
    >>> t[1,45] = 0
    >>> contains_no_value(t, 0)
    False

    >>> t = torch.full((100,100,), 0)
    >>> contains_no_value(t, 0)
    False
    """
    return ((tensor == torch.full_like(tensor, value)).float().sum() == 0).item()


from dataclasses import dataclass

from typing import Dict, Union, TypeVar


T = TypeVar('T')
TreeLikeDict = Dict[str, Union['TreeLikeDict', str]]


def patch_config(a: T, patch: Dict[str, str]) -> T:
    """
    >>> @dataclass(frozen=True)
    ... class C(object):
    ...     d: float = 3.0

    >>> @dataclass(frozen=True)
    ... class A(object):
    ...     a: int = 1
    ...     b: str = '2'
    ...     c: C = C()
    ...     d: bool = True

    >>> patch_config(A(), {})
    A(a=1, b='2', c=C(d=3.0), d=True)

    >>> patch_config(A(), {'c.d': '91', 'b': '89', 'd': 'False'})
    A(a=1, b='89', c=C(d=91.0), d=False)
    """
    tree_like_dct: TreeLikeDict = {}
    for path_to_param, new_value in patch.items():
        cur = tree_like_dct
        attrs = path_to_param.split(".")
        for attr in attrs[:-1]:
            if attr not in cur:
                cur[attr] = {}
            cur = cur[attr]
        cur[attrs[-1]] = new_value
    return patch_object(a, tree_like_dct)


def patch_object(a: T, params_to_be_patched: TreeLikeDict) -> T:
    """
    >>> @dataclass(frozen=True)
    ... class C(object):
    ...     d: float = 3.0

    >>> @dataclass(frozen=True)
    ... class A(object):
    ...     a: int = 1
    ...     b: str = '2'
    ...     c: C = C()
    ...     d: bool = True

    >>> patch_object(A(), {})
    A(a=1, b='2', c=C(d=3.0), d=True)

    >>> patch_object(A(), {'c': {'d': '91.0'}, 'b': '89', 'd': 'False'})
    A(a=1, b='89', c=C(d=91.0), d=False)
    """
    patched_attributes = {}
    for k, v in params_to_be_patched.items():
        attr_to_patch = getattr(a, k)
        if isinstance(v, dict):
            patched_attributes[k] = patch_object(attr_to_patch, v)
        else:
            tp = type(attr_to_patch)
            patched_attributes[k] = tp(v) if tp != bool else v.lower() == 'true'
    a_type = type(a)
    return a_type(**patched_attributes)



import jsons
from typing import Dict, Type, Union

from langmodels.lmconfig.datamodel import PrepFunction, LMTrainingConfig, LMTrainingMetrics, PrepFunctionOptions


def prep_function_serializer(prep_function: PrepFunction, **kwargs):
    return {'callable': prep_function.callable.__name__,
            'params': prep_function.params,
            'options':  jsons.dump(prep_function.options)}


def prep_function_deserializer(dct: Dict, cls: Type[PrepFunction], **kwargs) -> PrepFunction:
    import dataprep.api.corpus as api
    return cls(
        callable=getattr(api, dct['callable']),
        params=dct['params'],
        options=jsons.load(dct['options'], PrepFunctionOptions)
    )


jsons.set_serializer(prep_function_serializer, PrepFunction)
jsons.set_deserializer(prep_function_deserializer, cls=PrepFunction)


def dump_config_to_json(config: LMTrainingConfig) -> object:
    return jsons.dump(config)


def dump_config_to_string(config: LMTrainingConfig) -> str:
    return jsons.dumps(dump_config_to_json(config))


def dump_config(config: LMTrainingConfig, file: str) -> str:
    config_str = dump_config_to_string(config)
    with open(file, 'w') as f:
        f.write(config_str)
    return config_str


def load_config_from_string(s: str) -> Union[LMTrainingConfig, LMTrainingMetrics]:
    return jsons.loads(s, Union[LMTrainingConfig, LMTrainingMetrics], strict=True)


def load_config_from_file(file: str) -> Union[LMTrainingConfig, LMTrainingMetrics]:
    with open(file, 'r') as f:
        s = f.read().replace('\n', '')
    return load_config_from_string(s)


def read_value_from_file(file: str, value_type):
    with open(file, 'r') as f:
        res = f.read().rstrip('\n')
    return value_type(res)



import logging
import os
from pprint import pformat

import torch
from comet_ml import Experiment
from fastai.basic_data import DataBunch
from fastai.basic_train import validate
from fastai.callback import CallbackHandler, Callback
from fastai.callbacks.mem import PeakMemMetric
from fastai.callbacks.misc import StopAfterNBatches
from fastai.layers import FlattenedLoss
from fastai.metrics import accuracy
from fastai.text import Vocab, language_model_learner
from fastai.train import fit_one_cycle, Learner, EarlyStoppingCallback
from flatdict import FlatDict
from torch.nn import CrossEntropyLoss
from typing import Optional, Tuple

import dataprep.api.corpus as api
from dataprep.api.corpus import PreprocessedCorpus
from dataprep.util import to_literal_str
from langmodels.cuda_util import get_device_id
from langmodels.file_util import check_path_exists, check_path_writable, get_all_files
from langmodels.lmconfig.datamodel import LMTrainingConfig, Corpus, \
    RafaelsTrainingSchedule, TrainingProcedure, CosineLRSchedule, ExperimentRun, DeviceOptions
from langmodels.lmconfig.serialization import dump_config, dump_config_to_string, dump_config_to_json
from langmodels.model import TrainedModel, create_custom_config
from langmodels.modelregistry import load_from_path
from langmodels.tensor_ops import mrr
from langmodels.training.data import EmptyDataBunch, create_databunch
from langmodels.training.schedule import ReduceLRCallback
from langmodels.training.subepoch_files import EpochFileLoader
from langmodels.training.tracking import FirstModelTrainedCallback, LrLogger, RetryingSaveModelCalback

logger = logging.getLogger(__name__)

VOCAB_FILE_NAME = 'vocab'
CONFIG_FILE_NAME = 'config'

HOME = os.environ['HOME']
PATH_TO_PREP_DATASETS = os.environ['PATH_TO_PREP_DATASETS'] if 'PATH_TO_PREP_DATASETS' in os.environ \
    else os.path.join(HOME, 'prep-datasets')


def create_vocab_for_lm(prep_corpus: PreprocessedCorpus) -> Vocab:
    return Vocab(['`unk', '`pad'] + list(map(lambda x: to_literal_str(x), prep_corpus.load_vocab().keys())))


def choose_schedule_and_fit(learner: Learner, training_procedure: TrainingProcedure) -> None:
    schedule = training_procedure.schedule
    if isinstance(schedule, RafaelsTrainingSchedule):
        reduce_lr_callback = ReduceLRCallback(learner,
                                              mult_coeff=schedule.mult_coeff,
                                              max_times_lr_decrease=schedule.max_lr_reduction_times)
        learner.callbacks.append(reduce_lr_callback)
        learner.fit(epochs=schedule.max_epochs, lr=schedule.init_lr, wd=training_procedure.weight_decay)
    elif isinstance(schedule, CosineLRSchedule):
        if schedule.early_stop:
            learner.callbacks.append(EarlyStoppingCallback(learner, patience=schedule.early_stop.patience))
        fit_one_cycle(learner, cyc_len=schedule.cyc_len, tot_epochs=schedule.max_epochs,
                      max_lr=schedule.max_lr,
                      wd=training_procedure.weight_decay)
    # not saving the model explicitly because it should have been saved by the callbacks


def load_base_model_if_needed(learner: Learner, lm_training_config: LMTrainingConfig, model_file='best') -> None:
    if lm_training_config.base_model:
        model = os.path.join(lm_training_config.base_model, model_file)
        print(f"Using pretrained model: {model}.pth")
        # not setting purge to True raises a pickle serialization error
        learner.load(model, purge=False)
    else:
        print("Training form scratch")


def save_experiment_input(run: ExperimentRun, learner: Learner, vocab: Vocab):
    vocab.save(os.path.join(run.path_to_trained_model, VOCAB_FILE_NAME))
    dump_config(run.config, os.path.join(run.path_to_trained_model, CONFIG_FILE_NAME))
    if run.comet_experiment:
        save_params_to_comet(run.comet_experiment, run.config, vocab, get_param_number(learner))


def check_run_prerequisites(run: ExperimentRun) -> None:
    if run.config.base_model:
        check_path_exists(os.path.join(run.config.base_model, 'best.pth'))
    check_path_writable(run.path_to_trained_model)


def run_validation(trained_model: TrainedModel, corpus: Corpus, only_validation_files: bool = False,
                   fallback_to_cpu: bool = True, non_default_device_to_use: Optional[int] = None):
    """
    Validation using fastai's `validation` method
    """
    prep_corpus: api.PreprocessedCorpus = trained_model.prep_corpus(corpus)
    config: LMTrainingConfig = trained_model.config

    device_id = get_device_id(fallback_to_cpu, non_default_device_to_use)

    logger.info(f"Vocab size: {len(trained_model.vocab.itos)}")
    all_files = [f for f in get_all_files(prep_corpus.path_to_prep_dataset, None)]
    databunch = create_databunch(prep_corpus.path_to_prep_dataset, all_files, trained_model.vocab,
                                 bs=config.bs, bptt=config.bptt, device=device_id,
                                 only_validation_files=only_validation_files, allow_unks=True)

    class DetupleCallback(Callback):
        def on_loss_begin(self, last_output: Tuple[torch.Tensor, torch.Tensor, torch.Tensor], **kwargs):
            """Save the extra outputs for later and only returns the true output."""
            return {'last_output': last_output[0]}

    return validate(trained_model.model, databunch.valid_dl, loss_func=FlattenedLoss(CrossEntropyLoss),
                    cb_handler=CallbackHandler([DetupleCallback()]))


def get_param_number(learner: Learner) -> int:
    return sum(p.numel() for p in learner.model.parameters() if p.requires_grad)


def save_params_to_comet(experiment: Experiment, lm_training_config: LMTrainingConfig,
                         vocab: Vocab, trainable_params: int) -> Experiment:
    flat_config = FlatDict(dump_config_to_json(lm_training_config))
    for name, value in flat_config.items():
        experiment.log_parameter(name, value)
    experiment.log_parameter("vocabulary", len(vocab.itos))
    experiment.log_parameter("trainable_params", trainable_params)
    experiment.log_parameter("model_available", False)
    return experiment


def add_callbacks(experiment_run: ExperimentRun, learner: Learner, vocab: Vocab, tune: bool) -> None:
    comet_experiment: Optional[Experiment] = experiment_run.comet_experiment
    if comet_experiment:
        learner.callbacks.append(LrLogger(learner, comet_experiment))

    first_model_trained_callback = FirstModelTrainedCallback(learner, experiment_run)
    learner.callbacks.append(first_model_trained_callback)

    save_every_epoch_callback = RetryingSaveModelCalback(learner, every='epoch', name='epoch')
    learner.callbacks.append(save_every_epoch_callback)
    save_best_model_callback = RetryingSaveModelCalback(learner, every='improvement', name='best')
    learner.callbacks.append(save_best_model_callback)

    if tune:
        logger.warning("Tune mode is ON!")
        learner.callbacks.append(StopAfterNBatches(n_batches=2))


def train(training_config: LMTrainingConfig = LMTrainingConfig(),
          device_options: DeviceOptions() = DeviceOptions(),
          tune=False, comet=True) -> TrainedModel:
    logger.info(f'Using the following config: \n{pformat(dump_config_to_string(training_config))}')
    experiment_run = ExperimentRun.with_config(training_config, device_options=device_options, comet=comet)
    check_run_prerequisites(experiment_run)

    prep_corpus: api.PreprocessedCorpus = training_config.prep_function.apply(training_config.corpus,
                                                                              output_path=PATH_TO_PREP_DATASETS)
    vocab = create_vocab_for_lm(prep_corpus)
    print(f"Vocab size: {len(vocab.itos)}")
    device = get_device_id(device_options.fallback_to_cpu, device_options.non_default_device_to_use)
    empty_data_bunch: DataBunch = EmptyDataBunch(vocab=vocab, path=prep_corpus.path_to_prep_dataset, device=device)

    config = create_custom_config(training_config)
    arch_class = training_config.get_arch_class()
    learner = language_model_learner(empty_data_bunch, arch_class,
                                     drop_mult=training_config.arch.drop.multiplier,
                                     config=config, pretrained=not config, metrics=[accuracy, mrr],
                                     callback_fns=[PeakMemMetric] if torch.cuda.is_available() else [],
                                     path=os.path.dirname(experiment_run.path_to_trained_model),
                                     model_dir=os.path.basename(experiment_run.path_to_trained_model))

    files_per_epoch = training_config.training_procedure.files_per_epoch
    learner.callbacks.append(EpochFileLoader(learner, prep_corpus, vocab,
                                             bs=training_config.bs, bptt=training_config.bptt, device=device,
                                             n_files_per_epoch=files_per_epoch))

    save_experiment_input(experiment_run, learner, vocab)

    add_callbacks(experiment_run, learner, vocab, tune)

    load_base_model_if_needed(learner, training_config)

    print(f"Starting training... Model will be saved to {experiment_run.perm_path_to_model} "
          f"(Saving config and vocab to {experiment_run.path_to_trained_model} before getting the first trained model)")
    choose_schedule_and_fit(learner, training_config.training_procedure)

    # TODO export learner?
    return load_from_path(experiment_run.path_to_trained_model, force_use_cpu=True)



from typing import List
from unittest.mock import MagicMock


def file_mock_with_lines(lines: List[str]):
    file_mock = MagicMock(spec=['__enter__', '__exit__'])
    handle1 = file_mock.__enter__.return_value
    handle1.__iter__.return_value = iter(map(lambda l: l + '\n', lines))
    return file_mock


import hashlib
import logging
import os

import requests
from columnar import columnar
from typing import List

from langmodels import MODEL_ZOO_PATH
from langmodels.model import TrainedModel, ModelDescription

logger = logging.getLogger(__name__)

MODEL_DIR_URL = 'https://www.inf.unibz.it/~hbabii/pretrained_models'
MODEL_LIST_URL = MODEL_DIR_URL + '.list'


def _download_file(url: str, path: str, check_md5: bool = False):
    response = requests.get(url, stream=True)
    with open(path + '.tmp', "wb") as f:
        for chunk in response.iter_content(chunk_size=512):
            if chunk:
                f.write(chunk)

    if check_md5:
        correct_md5 = requests.get(url + ".md5").content.decode(encoding='utf8').rstrip('\n')
        md5 = hashlib.md5(open(path + '.tmp', 'rb').read()).hexdigest()
        if md5 != correct_md5:
            logger.warning(f'The md5 checksum for {url} is not correct: {md5}. Downloading the file again...')
            _download_file(url, path, check_md5=True)

    os.replace(path + '.tmp', path)


def _get_all_model_ids(cached: bool = False) -> List[str]:
    if cached:
        root, dirs, files = next(os.walk(MODEL_ZOO_PATH))
        return [d for d in dirs if os.path.exists(os.path.join(root, d, 'best.pth'))]
    else:
        content = requests.get(MODEL_LIST_URL).content.decode(encoding='utf8')
        model_list = content.rstrip('\n').split('\n')
        return model_list


def _get_all_models(cached: bool):
    model_ids = _get_all_model_ids(cached=cached)
    loaded_models = [load_model_by_id(model_id, load_description_only=True).get_model_description()
                     for model_id in model_ids]
    return loaded_models


class _ModelQuery(object):
    def __init__(self, previous_query: '_ModelQuery'):
        self.previous_query = previous_query

    def __str__(self):
        desc_list = self.execute()
        return columnar(headers=ModelDescription.get_attribute_list(),
                        data=list(map(lambda l: l.get_value_list(), desc_list)),
                        no_borders=True, terminal_width=200)

    def sorted_by_entropy(self) -> '_ModelQuery':
        return _SortByEntropyQuery(self)

    def get_previous_query(self) -> '_ModelQuery':
        return self.previous_query


class _SortByEntropyQuery(_ModelQuery):
    def __init__(self, previous_query: _ModelQuery):
        super().__init__(previous_query)

    def execute(self) -> List[ModelDescription]:
        return sorted(self.get_previous_query().execute(), key=lambda m: m.bin_entropy)


class _GetAllModelsQuery(_ModelQuery):
    def __init__(self, cached: bool):
        super().__init__(None)
        self.cached = cached

    def execute(self) -> List[ModelDescription]:
        return _get_all_models(cached=self.cached)


def _get_all_models_query(cached: bool) -> _ModelQuery:
    return _GetAllModelsQuery(cached=cached)


def query_all_models(cached: bool = False) -> List[ModelDescription]:
    return _get_all_models_query(cached=cached).sorted_by_entropy().execute()


def list_pretrained_models(cached: bool = False) -> None:
    print(_get_all_models_query(cached=cached).sorted_by_entropy())


def load_from_path(path: str, force_use_cpu: bool = False, load_description_only: bool = False) -> TrainedModel:
    return TrainedModel(path, force_use_cpu, load_description_only)


def load_model_by_id(id: str, force_use_cpu: bool = False, load_description_only: bool = False) -> TrainedModel:
    path = os.path.join(MODEL_ZOO_PATH, id)
    if not os.path.exists(os.path.join(path, 'best.pth')):
        if id not in _get_all_model_ids():
            raise ValueError(f'Model with id {id} is not found on the server.')
        url_to_model_dir = MODEL_DIR_URL + '/' + id
        if not os.path.exists(path):
            os.makedirs(path)
        _download_file(url_to_model_dir + '/config', os.path.join(path, 'config'))
        _download_file(url_to_model_dir + '/metrics', os.path.join(path, 'metrics'))
        _download_file(url_to_model_dir + '/tags', os.path.join(path, 'tags'))
        if not load_description_only:
            logger.info(f'Model is not found in cache. Downloading from {url_to_model_dir} ...')
            _download_file(url_to_model_dir + '/best.pth', os.path.join(path, 'best.pth'), check_md5=True)
            _download_file(url_to_model_dir + '/vocab', os.path.join(path, 'vocab'), check_md5=True)

    return load_from_path(path, force_use_cpu, load_description_only)


def load_model_with_tag(tag: str, force_use_cpu: bool = False):
    for cached in [True, False]:  # first checking tags of cached models
        for model in query_all_models(cached=cached):
            if model.is_tagged_by(tag):
                return load_model_by_id(model.id, force_use_cpu=force_use_cpu)
    raise ValueError(f'Model tagged with {tag} not found.')


def load_default_model(force_use_cpu: bool = False) -> TrainedModel:
    return load_model_with_tag('DEFAULT', force_use_cpu=force_use_cpu)


__all__ = [load_default_model, list_pretrained_models, query_all_models, load_model_by_id, load_model_with_tag, load_from_path]



import logging
import os

from fastai.basic_train import LearnerCallback, Learner
from fastai.callback import Callback
from fastai.text import Vocab
from pathlib import Path
from typing import Any, Optional, List, Tuple

from dataprep.api.corpus import PreprocessedCorpus
from langmodels.file_util import get_all_files
from langmodels.training.data import create_databunch, check_data

BIG_EPOCH_FILE_LIMIT = 10 * 1000

TRAIN_SUBDIR = 'train'
TEST_SUBDIR = 'test'
VALID_SUBDIR = 'valid'


logger = logging.getLogger(__name__)


class EpochFileLoader(LearnerCallback):
    def __init__(self, learner: Learner, prep_corpus: PreprocessedCorpus,
                 vocab: Vocab, bs: int, bptt: int, device: str, n_files_per_epoch: Optional[int]):
        super().__init__(learner)

        if n_files_per_epoch is not None and n_files_per_epoch <= 0:
            raise ValueError(f'The value of n_files_per_epoch should be > 0 or None but is {n_files_per_epoch}')

        self.prep_corpus = prep_corpus
        self.vocab: Vocab = vocab
        self.bs: int = bs
        self.bptt: int = bptt
        self.device: str = device
        self.n_files_per_epoch = n_files_per_epoch

        self.path_to_train_files = os.path.join(self.prep_corpus.path_to_prep_dataset, TRAIN_SUBDIR)
        self.valid_and_test_files = self._load_valid_and_test_files()
        self.train_files_iterator = self._reset_train_file_iterator()

    def _reset_train_file_iterator(self):
        return get_all_files(self.path_to_train_files, None)

    def _handle_empty_training_files_list(self, **kwargs):
        if 'after_empty_epoch' in kwargs:
            raise ValueError("Looks like your training set is empty.")
        else:
            self.on_epoch_begin(after_empty_epoch=True, **kwargs)

    def on_epoch_begin(self, **kwargs: Any) -> None:
        train_files, all_files_read = self._get_next_training_files(self.n_files_per_epoch)
        if all_files_read:
            self.train_files_iterator = self._reset_train_file_iterator()
        if len(train_files) == 0:
            self._handle_empty_training_files_list(**kwargs)
            return
        small_epoch = len(train_files) < BIG_EPOCH_FILE_LIMIT
        if self.n_files_per_epoch is not None:
            train_subfolders = {f.relative_to(self.path_to_train_files).parts[0] for f in train_files}
            logger.info(f"Using projects for training: {','.join(train_subfolders)}")
        databunch = create_databunch(self.prep_corpus.path_to_prep_dataset, self.valid_and_test_files + train_files,
                                     self.vocab, bs=self.bs, bptt=self.bptt,
                                     device=self.device, verbose=not small_epoch)
        check_data(databunch, self.vocab, verbose=not small_epoch)
        self.learn.data = databunch
        train_dl = self.learn.data.train_dl
        if isinstance(train_dl, Callback):
            train_dl.on_train_begin()

    def _load_valid_and_test_files(self) -> List[Path]:
        path_to_valid_files = os.path.join(self.prep_corpus.path_to_prep_dataset, VALID_SUBDIR)
        path_to_test_files = os.path.join(self.prep_corpus.path_to_prep_dataset, TEST_SUBDIR)

        return [f for f in get_all_files(path_to_valid_files, None)] + \
               [f for f in get_all_files(path_to_test_files, None)]

    def _get_next_training_files(self, n: Optional[int]) -> Tuple[List[Path], bool]:
        res: List[Path] = []
        for file in self.train_files_iterator:
            res.append(file)
            if n and n == len(res):
                return res, False
        return res, True



import logging
from dataclasses import dataclass

import numpy as np
from typing import List, Tuple, Optional, Set, Callable, Dict, Type

from dataprep.parse.model.metadata import PreprocessingMetadata
from langmodels.evaluation.filtering import FullTokenIterator, to_full_token_string, FilteringTokenIterator, \
    TokenIterator, SubtokenIterator, MetricName, TokenTypes
from langmodels.model import TrainedModel

DEFAULT_N_MODEL_SUGGESTIONS = 100


logger = logging.getLogger(__name__)

InclusionFunction = Callable[[PreprocessingMetadata, int], bool]
AverageFunction = Callable[[List[float], PreprocessingMetadata, InclusionFunction], Tuple[float, int]]


def conditional_average(subword_entropies: List[float],
                        metadata: PreprocessingMetadata,
                        inclusion_function: InclusionFunction = lambda m, i: True,
                        token_iterator_type: Type[TokenIterator] = FullTokenIterator) \
        -> Tuple[float, int]:
    """evaluator
    >>> conditional_average([], PreprocessingMetadata(set(), [0]))
    (0.0, 0)

    >>> conditional_average([1.0, 2.0, 3.0], PreprocessingMetadata(set(), word_boundaries=[0, 1, 3]))
    (3.0, 2)

    >>> metadata = PreprocessingMetadata(set(), word_boundaries=[0, 1, 3], comments=[(0, 1)])
    >>> conditional_average([1.0, 2.0, 3.0], metadata, inclusion_func_dict[TokenTypes.ONLY_COMMENTS])
    (1.0, 1)

    >>> metadata = PreprocessingMetadata(set(), word_boundaries=[0, 1, 3], comments=[(0, 1)])
    >>> conditional_average([1.0, 2.0, 3.0], metadata, inclusion_func_dict[TokenTypes.ALL_BUT_COMMENTS])
    (5.0, 1)

    >>> metadata = PreprocessingMetadata(set(), word_boundaries=[0, 1, 3], comments=[(0, 1)])
    >>> conditional_average([1.0, 2.0, 3.0], metadata, inclusion_func_dict[TokenTypes.ALL_BUT_COMMENTS], SubtokenIterator)
    (2.5, 2)
    """
    word_entropies = [we for we in FilteringTokenIterator(subword_entropies, metadata, inclusion_function,
                                                          token_iterator_type=token_iterator_type, agg=sum)]
    if not word_entropies:
        return .0, 0

    n_full_tokens = len(word_entropies)
    return sum(word_entropies) / n_full_tokens, n_full_tokens


inclusion_func_dict: Dict[TokenTypes, InclusionFunction] = {
    TokenTypes.ALL: lambda m, i: True,
    TokenTypes.ONLY_COMMENTS: lambda m, i: m.is_comment_at_index(i),
    TokenTypes.ALL_BUT_COMMENTS: lambda m, i: not m.is_comment_at_index(i)
}


@dataclass(frozen=True)
class EvaluationResult(object):
    subtoken_values: List[Optional[float]]  # this value should correspond to the number of subtokens
    average: float
    n_samples: int


@dataclass(frozen=True)
class EvaluationScenario(object):
    metric_name: MetricName
    token_types: TokenTypes

    def __str__(self):
        return f'{self.metric_name}/{self.token_types}'

    def __repr__(self):
        return str(self)


@dataclass(frozen=True)
class Evaluation(object):
    text: str
    prep_text: List[str]
    prep_metadata: PreprocessingMetadata
    scenarios: Dict[EvaluationScenario, EvaluationResult]


def bin_entropy(model: TrainedModel, prep_line: List[str], prep_metadata: PreprocessingMetadata,
                token_types: Optional[Set[TokenTypes]] = None,
                token_iterator_types: Set[Type[TokenIterator]] = None) -> Dict[EvaluationScenario, EvaluationResult]:
    """
    Changes the state of the model!
    """
    token_types = token_types or [TokenTypes.ALL]
    token_iterator_types = token_iterator_types or {FullTokenIterator}

    iterator_to_metric_map = {FullTokenIterator: 'full_token_entropy', SubtokenIterator: 'subtoken_entropy'}

    all_entropies = model.get_entropies_for_prep_text(prep_line)
    evaluation_results = {}
    for tt in token_types:
        for token_iterator_type in token_iterator_types:
            filtered_entropies = []
            inclusion_func = inclusion_func_dict[tt]
            for ind, e in enumerate(all_entropies):
                filtered_entropies.append(e if inclusion_func(prep_metadata, ind) else None)
            aggregated_entropy, n_full_tokens = conditional_average(filtered_entropies, prep_metadata,
                                                                    inclusion_func, token_iterator_type)
            evaluation_results[EvaluationScenario(iterator_to_metric_map[token_iterator_type], tt)] = \
                EvaluationResult(filtered_entropies, aggregated_entropy, n_full_tokens)
    return evaluation_results


def mrr(model: TrainedModel,
        prep_line: List[str],
        metadata: PreprocessingMetadata,
        token_types: Optional[Set[TokenTypes]] = None) -> Dict[EvaluationScenario, EvaluationResult]:
    """
    Changes the state of the model!
    """
    token_types = token_types or [TokenTypes.ALL]
    result: Dict[EvaluationScenario, EvaluationResult] = {}
    for tt in token_types:
        incl_func = inclusion_func_dict[tt]
        inverse_rank_sum = .0
        count = 0
        inverse_ranks: List[Optional[float]] = []
        for full_token in FilteringTokenIterator(prep_line, metadata, lambda m, i: True, agg=lambda l:l):
            predictions = model.predict_next_full_token(n_suggestions=DEFAULT_N_MODEL_SUGGESTIONS)
            predicted_tokens = list(map(lambda p: p[0], predictions))
            actual_token = to_full_token_string(full_token)
            if incl_func(metadata, len(inverse_ranks)):
                try:
                    rank = predicted_tokens.index(actual_token) + 1
                    inverse_rank = 1. / rank
                except ValueError:  # actual token is not in prediction list
                    inverse_rank = 0.
                inverse_rank_sum += inverse_rank
                inverse_ranks.append(inverse_rank)
                inverse_ranks.extend([None for i in range(len(full_token) - 1)])
                count += 1
            else:
                inverse_ranks.extend([None for i in range(len(full_token))])
            model.feed_text(actual_token)
        result[EvaluationScenario('mrr', tt)] = EvaluationResult(inverse_ranks, inverse_rank_sum / count if count else 0., count)

    return result


def average_ranks(values: List[float], weights: List[int] = None) -> Tuple[float, int]:
    """
    >>> average_ranks([1, 5])
    (1.6666666666666667, 2)

    >>> average_ranks([1, 5, 9])
    (2.2881355932203387, 3)

    >>> average_ranks([1.6666666666666667, 9], weights=[2, 1])
    (2.2881355932203387, 3)
    """
    weights = weights or [1.0] * len(values)
    return 1.0 / np.average(list(map(lambda x: 1 / x, values)), weights=weights), int(sum(weights))


def average_entropy(values: List[float], weights: List[int] = None) -> Tuple[float, int]:
    """
    >>> average_entropy([6, 9], weights=[2, 1])
    (7.0, 3)
    """
    return np.average(values, weights=weights), sum(weights)


Metric = Callable[[TrainedModel, List[str], PreprocessingMetadata, Optional[Set[TokenTypes]], Type[TokenIterator]],
                  Dict[EvaluationScenario, EvaluationResult]]
MetricAggregator = Callable[[List[float], List[int]], Tuple[float, int]]

metric_dict: Dict[MetricName, Tuple[Metric, MetricAggregator]] = {
    'subtoken_entropy': (lambda m, sw, md, tt: bin_entropy(m, sw, md, tt, {SubtokenIterator}), average_entropy),
    'full_token_entropy': (bin_entropy, average_entropy),
    'mrr': (mrr, average_ranks)
}


def get_metric_func_by_name(name: str) -> Metric:
    return metric_dict[name][0]


def get_metric_aggregator_by_name(name: str) -> MetricAggregator:
    return metric_dict[name][1]


def entropy_to_probability(entropy: float) -> float:
    """
    >>> entropy_to_probability(0.0)
    1.0

    >>> entropy_to_probability(1.0)
    0.5

    >>> entropy_to_probability(3.0)
    0.125

    >>> entropy_to_probability(100.0)
    7.888609052210118e-31
    """
    return 2 ** -entropy





import os
import sys
from dataclasses import dataclass, field, asdict

from comet_ml import Experiment
from typing import Optional, Callable, Tuple, Union, List, Type

import dataprep.api.corpus as corpus_api
from dataprep.api.corpus import PreprocessedCorpus
from fastai.text import AWD_LSTM, Transformer, TransformerXL, Activation

from langmodels import MODEL_ZOO_PATH, __version__, __major_version__
from langmodels.nn import GRU

CONFIG_VERSION = __major_version__ if __major_version__ > 0 else __version__

HOME = os.environ['HOME']

TMP_SUFFIX = '.tmp'


@dataclass(frozen=True)
class Dropouts(object):
    multiplier: float = 0.5
    oute: float = 0.02
    outi: float = 0.25
    outh: float = 0.15
    w: float = 0.2
    out: float = 0.1


@dataclass(frozen=True)
class RegFn(object):
    alpha: float = 2.
    beta: float = 1.


@dataclass(frozen=True)
class TrainingSchedule(object):
    pass


@dataclass(frozen=True)
class RafaelsTrainingSchedule(TrainingSchedule):
    init_lr: float = 1e-4
    mult_coeff: float = 0.5
    max_epochs: int = 50
    max_lr_reduction_times: int = 6


@dataclass(frozen=True)
class EarlyStop(object):
    patience: int = 3


@dataclass(frozen=True)
class CosineLRSchedule(TrainingSchedule):
    max_lr: float = 1e-4
    cyc_len: int = 3
    max_epochs: int = 30
    early_stop: EarlyStop = EarlyStop()


@dataclass(frozen=True)
class Corpus(object):
    path: str = os.path.join(HOME, 'dataset')
    extensions: str = 'java'  # in format "py" or "java|c|py"


PrepCallable = Callable[..., PreprocessedCorpus]
ParametrizedPrepCallable = Callable[[Corpus], PreprocessedCorpus]


@dataclass(frozen=True)
class PrepFunctionOptions(object):
    no_unicode: bool = True
    no_spaces: bool = True
    no_com: bool = False
    no_str: bool = False
    max_str_length: int = sys.maxsize


@dataclass(frozen=True)
class PrepFunction(object):
    callable: PrepCallable = corpus_api.bpe
    params: List[str] = field(default_factory=lambda: ['10k'])
    options: PrepFunctionOptions = PrepFunctionOptions()

    @property
    def apply(self) -> ParametrizedPrepCallable:
        def prep_corpus(corpus: Corpus, **kwargs) -> PreprocessedCorpus:
            return self.callable(corpus.path, *self.params, **asdict(self.options), **kwargs,
                                 extensions=corpus.extensions)

        return prep_corpus


@dataclass(frozen=True)
class LstmArch(object):
    bidir: bool = False
    qrnn: bool = False
    emb_sz: int = 1024
    n_hid: int = 1024
    n_layers: int = 3
    adam_betas: Tuple[float, float] = (0.7, 0.99)
    clip: float = 0.3
    reg_fn: RegFn = RegFn()
    drop: Dropouts = Dropouts()
    tie_weights: bool = True
    out_bias: bool = True
    lstm: bool = True

    def __post_init__(self):
        if not self.lstm:
            raise TypeError(f'Value of lstm field in LstmArch must be set to True')


@dataclass(frozen=True)
class GruArch(object):
    bidir: bool = False
    emb_sz: int = 1024
    n_hid: int = 1024
    n_layers: int = 3
    adam_betas: Tuple[float, float] = (0.7, 0.99)
    clip: float = 0.3
    reg_fn: RegFn = RegFn()
    drop: Dropouts = Dropouts()
    tie_weights: bool = True
    out_bias: bool = True
    gru: bool = True

    def __post_init__(self):
        if not self.gru:
            raise TypeError(f'Value of gru field in GruArch must be set to True')


@dataclass(frozen=True)
class TransformerDropouts(object):
    multiplier: float = 1.0
    resid: float = 0.1
    attn: float = 0.1
    ff: float = 0.1
    embed: float = 0.1
    output: float = 0.


@dataclass(frozen=True)
class TransformerArch(object):
    ctx_len: int = 256
    n_layers: int = 3
    n_heads: int = 6
    d_model: int = 512
    d_head: int = 16
    d_inner: int = 2048
    drop: TransformerDropouts = TransformerDropouts()
    bias: bool = True
    scale: bool = True
    act: Activation = Activation.GeLU
    double_drop: bool = False
    tie_weights: bool = True
    out_bias: bool = False
    mask: bool = True


@dataclass(frozen=True)
class TrainingProcedure(object):
    schedule: Union[CosineLRSchedule, RafaelsTrainingSchedule] = RafaelsTrainingSchedule()
    files_per_epoch: Optional[int] = 50 * 1000
    weight_decay: float = 1e-6


@dataclass(frozen=True)
class DeviceOptions(object):
    fallback_to_cpu: bool = False
    non_default_device_to_use: Optional[int] = None


@dataclass(frozen=True)
class LMTrainingConfig(object):
    corpus: Corpus = Corpus()
    base_model: Optional[str] = None
    bs: int = 32
    prep_function: PrepFunction = PrepFunction()
    arch: Union[LstmArch, GruArch, TransformerArch] = LstmArch()
    bptt: int = 200
    training_procedure: TrainingProcedure = TrainingProcedure()
    config_version: str = CONFIG_VERSION

    def __post_init__(self):
        if self.config_version != CONFIG_VERSION:
            raise TypeError(f'Trying to deserealize '
                            f'CONFIG_VERSION {self.config_version} '
                            f'to CONFIG_VERSION {CONFIG_VERSION} object')

    def get_arch_class(self) -> Union[Type[AWD_LSTM], Type[GRU], Type[Transformer]]:
        if isinstance(self.arch, LstmArch):
            return AWD_LSTM
        elif isinstance(self.arch, GruArch):
            return GRU
        elif isinstance(self.arch, TransformerArch):
            return Transformer
        else:
            raise ValueError(f"Unknown architecture: {self.arch}")


def create_comet_experiment(run_id: str):
    experiment = Experiment()
    experiment.set_name(run_id)
    return experiment


class ExperimentRun:
    def __init__(self, run_id: str, config: LMTrainingConfig, device_options: DeviceOptions, comet_experiment: Optional):
        self.id = run_id
        self.config = config
        self.gpu = device_options
        self.comet_experiment: Optional[Experiment] = comet_experiment
        self.first_model_trained = False

    @classmethod
    def with_config(cls, config: LMTrainingConfig, device_options: DeviceOptions = DeviceOptions(), comet: bool = True):
        run_id = ExperimentRun._generate_run_id(config)
        comet_experiment = create_comet_experiment(run_id) if comet else None
        return cls(run_id, config, device_options, comet_experiment)

    def set_first_model_trained(self):
        self.first_model_trained = True

    @staticmethod
    def _generate_run_id(config: LMTrainingConfig) -> str:
        name_parts = []
        if config.base_model:
            name_parts.append([os.path.basename(config.base_model)])

        dataset = os.path.basename(config.corpus.path)
        prep_func_param = config.prep_function.params[0]
        n_layers = config.arch.n_layers
        n_hid = config.arch.n_hid if not isinstance(config.arch, TransformerArch) \
            else config.arch.d_inner

        import datetime
        time_now = datetime.datetime.now()
        timestamp = f"{time_now:%y%m%d.%H%M%S}"

        name_parts.append([dataset, str(prep_func_param), str(n_layers), str(n_hid), timestamp])

        return "_-_".join(map(lambda p: "_".join(p), name_parts))

    @property
    def path_to_trained_model(self) -> str:
        return self.perm_path_to_model if self.first_model_trained else f'{self.perm_path_to_model}{TMP_SUFFIX}'

    @property
    def perm_path_to_model(self) -> str:
        return os.path.join(MODEL_ZOO_PATH, self.id)


@dataclass
class LMTrainingMetrics(object):
    bin_entropy: float
    training_time_minutes_per_epoch: int
    n_epochs: int
    best_epoch: int



import logging

import torch
from typing import Union, Optional

from torch import cuda


logger = logging.getLogger(__name__)


class CudaNotAvailable(Exception):
    pass


# TODO these two methods below probably can be merged
def get_device(force_use_cpu: bool = False) -> Union[int, str]:
    return cuda.current_device() if cuda.is_available() and not force_use_cpu else 'cpu'


def get_device_id(fallback_to_cpu: bool = True, non_default_device_to_use: Optional[int] = None) \
        -> Union[str, torch.device]:
    if cuda.is_available():
        device_id = non_default_device_to_use or cuda.current_device()
        return torch.device('cuda', device_id)
    elif fallback_to_cpu:
        return "cpu"
    else:
        raise CudaNotAvailable("Cuda not available")


def get_map_location(force_use_cpu: bool):
    if force_use_cpu:
        map_location = lambda storage, loc: storage
        logger.debug("Using CPU for inference")
    elif cuda.is_available():
        map_location = torch.device('cuda:0')
        logger.debug("Using GPU for inference")
    else:
        map_location = lambda storage, loc: storage
        logger.info("Cuda not available. Falling back to using CPU.")
    return map_location



from dataprep.parse.model.placeholders import placeholders
from fastai.text import Vocab

from langmodels.model import _create_term_vocab

cpe = placeholders['compound_word_end']


def test_simple_vocab():
    input_vocab = Vocab(['a', f'b{cpe}', 'c', 'd', f'e{cpe}'])
    actual_vocab, actual_first_non_term_index = _create_term_vocab(input_vocab)

    expected_itos = [f'b{cpe}', f'e{cpe}', 'a', 'c', 'd']

    assert expected_itos == actual_vocab.itos
    assert 2 == actual_first_non_term_index


def test_empty_vocab():
    input_vocab = Vocab([])
    actual_vocab, actual_first_non_term_index = _create_term_vocab(input_vocab)

    expected_itos = []

    assert expected_itos == actual_vocab.itos
    assert 0 == actual_first_non_term_index


def test_only_terminals():
    input_vocab = Vocab([f'b{cpe}', f'e{cpe}'])
    actual_vocab, actual_first_non_term_index = _create_term_vocab(input_vocab)

    expected_itos = [f'b{cpe}', f'e{cpe}']

    assert expected_itos == actual_vocab.itos
    assert 2 == actual_first_non_term_index


def test_only_non_terminals():
    input_vocab = Vocab(['a', 'c', 'd'])
    actual_vocab, actual_first_non_term_index = _create_term_vocab(input_vocab)

    expected_itos = ['a', 'c', 'd']

    assert expected_itos == actual_vocab.itos
    assert 0 == actual_first_non_term_index



from pytest import fixture
from pytest_mock.plugin import MockFixture

import langmodels.training.cli as cli
from langmodels.lmconfig.datamodel import DeviceOptions, LMTrainingConfig
from langmodels.training.__main__ import run


@fixture
def train_func_mocker(mocker: MockFixture):
    mocker.patch('langmodels.training.cli.train')
    return mocker


def test_defaults(train_func_mocker):
    argv = ['train']
    run(argv)
    cli.train.assert_called_with(tune=False, comet=True,
                                 device_options=DeviceOptions(fallback_to_cpu=False, non_default_device_to_use=0),
                                 training_config=LMTrainingConfig())


def test_device_comet_cpu(train_func_mocker):
    argv = ['train', '--fallback-to-cpu', '--tune', '--disable-comet', '--device=3']
    run(argv)
    cli.train.assert_called_with(tune=True, comet=False,
                                 device_options=DeviceOptions(fallback_to_cpu=True, non_default_device_to_use=3),
                                 training_config=LMTrainingConfig())


def test_short_options(train_func_mocker):
    argv = ['train', '-Cxt', '-d 3']
    run(argv)
    cli.train.assert_called_with(tune=True, comet=False,
                                 device_options=DeviceOptions(fallback_to_cpu=True, non_default_device_to_use=3),
                                 training_config=LMTrainingConfig())



import numpy as np
from fastai.text import Vocab, TextList
from pathlib import Path
from unittest.mock import patch

from langmodels.training.numericalize import Numericalizer
from tests.file_test_util import file_mock_with_lines


def all_trues(bools):
    try:
        fltn_bools = np.hstack(bools)
    except: # should not pass silently.
        fltn_bools = np.array(all_trues(a) for a in bools)
    ints = fltn_bools.prod()
    if isinstance(ints, np.ndarray):
        return all_trues(ints)
    return bool(ints)


@patch('langmodels.training.training.open')
def test_simple(open_mock):
    # Given
    file_mock1 = file_mock_with_lines(['1', 'My Class'])
    file_mock2 = file_mock_with_lines(['1', 'hi'])

    open_mock.side_effect = [file_mock1, file_mock2]

    numericalizer = Numericalizer(Vocab(['`unk', '`pad', '1', 'My', 'Class', 'hi']), n_cpus=1)

    text_list = TextList([Path('/path/to/some/file1'), Path('/path/to/some/file2')])

    # when
    numericalizer.process(text_list)

    # then
    expected = np.array([np.array([2, 3, 4]), np.array([2, 5])])
    assert all_trues(np.equal(expected, text_list.items, dtype=np.object))



from dataprep.parse.model.metadata import PreprocessingMetadata
from langmodels.evaluation.filtering import TokenTypes, FullTokenIterator, \
    SubtokenIterator

from langmodels.evaluation.metrics import bin_entropy, mrr, EvaluationResult, EvaluationScenario
from unittest.mock import Mock

from langmodels.model import TrainedModel


def test_bin_entropy_empty():
    trained_model_mock = Mock(spec=TrainedModel)
    trained_model_mock.get_entropies_for_prep_text.return_value = []

    expected = {EvaluationScenario('full_token_entropy', TokenTypes.ALL): EvaluationResult([], 0.0, 0)}
    actual = bin_entropy(trained_model_mock, [], PreprocessingMetadata())

    assert expected == actual


def test_bin_entropy_simple_args():
    trained_model_mock = Mock(spec=TrainedModel)
    trained_model_mock.get_entropies_for_prep_text.return_value = [1.0, 2.0]

    expected = {EvaluationScenario('full_token_entropy', TokenTypes.ALL): EvaluationResult([1.0, 2.0], 3.0, 1)}
    actual = bin_entropy(trained_model_mock, ['My', 'Class</t>'], PreprocessingMetadata(word_boundaries=[0, 2]))

    assert expected == actual


def test_bin_entropy_with_comment():
    trained_model_mock = Mock(spec=TrainedModel)
    trained_model_mock.get_entropies_for_prep_text.return_value = [1.0, 2.0, 3.0, 6.0]

    expected = {
        EvaluationScenario('full_token_entropy', TokenTypes.ALL):
            EvaluationResult([1.0, 2.0, 3.0, 6.0], 4.0, 3),

        EvaluationScenario('full_token_entropy', TokenTypes.ONLY_COMMENTS):
            EvaluationResult([None, None, 3.0, 6.0], 4.5, 2),

        EvaluationScenario('full_token_entropy', TokenTypes.ALL_BUT_COMMENTS):
            EvaluationResult([1.0, 2.0, None, None], 3.0, 1),

        EvaluationScenario('subtoken_entropy', TokenTypes.ALL):
            EvaluationResult([1.0, 2.0, 3.0, 6.0], 3.0, 4),

        EvaluationScenario('subtoken_entropy', TokenTypes.ONLY_COMMENTS):
            EvaluationResult([None, None, 3.0, 6.0], 4.5, 2),

        EvaluationScenario('subtoken_entropy', TokenTypes.ALL_BUT_COMMENTS):
            EvaluationResult([1.0, 2.0, None, None], 1.5, 2)
    }

    actual = bin_entropy(trained_model_mock, ['My', 'Class</t>', '/', '/'],
                         PreprocessingMetadata(word_boundaries=[0, 2, 3, 4], comments=[(2, 4)]),
                         token_types={TokenTypes.ALL, TokenTypes.ONLY_COMMENTS, TokenTypes.ALL_BUT_COMMENTS},
                         token_iterator_types={FullTokenIterator, SubtokenIterator})

    assert expected == actual


def test_mrr_default_args():
    trained_model_mock = Mock(spec=TrainedModel)
    trained_model_mock.predict_next_full_token.side_effect = [[('a1', 0.), ('b1', 0.)],
                                                              [('a2', 0.), ('b2', 0.)]]

    expected = {EvaluationScenario('mrr', TokenTypes.ALL): EvaluationResult([1.0, 0.5], 0.75, 2)}
    actual = mrr(trained_model_mock, ['a1</t>', 'b2</t>'], metadata=PreprocessingMetadata(word_boundaries=[0, 1, 2]))

    assert actual == expected


def test_mrr_default_all_token_types():
    trained_model_mock = Mock(spec=TrainedModel)
    trained_model_mock.predict_next_full_token.side_effect = [[('a1', 0.), ('b1', 0.)],
                                                              [('a2', 0.), ('b2', 0.)],
                                                              [('a3', 0.), ('b3', 0.)],
                                                              [('a4', 0.), ('b4', 0.)],

                                                              [('a1', 0.), ('b1', 0.)],
                                                              [('a2', 0.), ('b2', 0.)],
                                                              [('a3', 0.), ('b3', 0.)],
                                                              [('a4', 0.), ('b4', 0.)],

                                                              [('a1', 0.), ('b1', 0.)],
                                                              [('a2', 0.), ('b2', 0.)],
                                                              [('a3', 0.), ('b3', 0.)],
                                                              [('a4', 0.), ('b4', 0.)]
                                                              ]

    expected = {
        EvaluationScenario('mrr', TokenTypes.ALL):
            EvaluationResult([1.0, None, 0.5, None, 0., 0.], 0.375, 4),
        EvaluationScenario('mrr', TokenTypes.ONLY_COMMENTS):
            EvaluationResult([None, None, None, None, 0., 0.], 0., 2),
        EvaluationScenario('mrr', TokenTypes.ALL_BUT_COMMENTS):
            EvaluationResult([1.0, None, 0.5, None, None, None], 0.75, 2)
    }
    actual = mrr(trained_model_mock, ['a', '1</t>', 'b', '2</t>', '/</t>', '/</t>'],
                 metadata=PreprocessingMetadata(word_boundaries=[0, 2, 4, 5, 6], comments=[(4, 6)]),
                 token_types={TokenTypes.ALL, TokenTypes.ONLY_COMMENTS, TokenTypes.ALL_BUT_COMMENTS})

    assert actual == expected




import logging
import os
import typing
from collections import Counter
from concurrent.futures import ProcessPoolExecutor
from functools import reduce
from typing import Collection, List, Tuple

import numpy as np
from dataprep.util import merge_dicts_
from fastai.core import partition_by_cores
from fastai.data_block import PreProcessor
from fastai.text import Vocab
from tqdm import tqdm


logger = logging.getLogger(__name__)


class Numericalizer(PreProcessor):
    def __init__(self, vocab: Vocab, n_cpus: int = os.cpu_count(),
                 allow_unks: bool = False, large_databunch: bool = False):
        super().__init__()
        self.vocab = vocab
        self.n_cpus = n_cpus
        self.allow_unks = allow_unks
        self.large_databunch = large_databunch

    def process_one(self, item: str):
        with open(item, 'r') as f:
            prep_tokens = [token for line in f for token in line.rstrip('\n').split(' ')]
            unks = Counter()
            for prep_token in prep_tokens:
                if prep_token not in self.vocab.itos:
                    if self.allow_unks:
                        unks.update([prep_token])
                    else:
                        raise ValueError(f'{[prep_token]} is not present in the vocabulary.\n'
                                         f'Vocab size is {len(self.vocab.itos)}. Vocab is {self.vocab.itos}')
            numericalized_tokens = np.array(self.vocab.numericalize(prep_tokens), dtype=np.int64)
            return numericalized_tokens, unks

    def process(self, ds: Collection) -> None:
        ds.vocab = self.vocab
        tokens, unks = self._process_in_parallel(ds.items)
        ds.items = np.array(tokens)
        if self.allow_unks:
            logger.warning(f"Encountered the following unknown tokens "
                           f"{sorted(unks.items(), reverse=True, key=lambda x:x[1])}")

    def _process_on_one_core(self, items: Collection[str]) -> List[Tuple[np.ndarray, typing.Mapping[str, int]]]:
        items = tqdm(items) if self.large_databunch else items
        return [self.process_one(item) for item in items]

    def _process_in_parallel(self, texts: Collection[str]) -> Tuple[List[np.ndarray], typing.Mapping[str, int]]:
        if self.n_cpus <= 1:
            if len(texts) == 0:
                return [[]], Counter()
            all_tokens, unk_list = zip(*self._process_on_one_core(texts))
            return list(all_tokens), reduce(lambda x, y: merge_dicts_(x, y)[0], unk_list, {})
        with ProcessPoolExecutor(self.n_cpus) as e:
            all_tokens = []
            all_unks = {}
            for from_one_core in e.map(self._process_on_one_core, partition_by_cores(texts, self.n_cpus)):
                for tokens, unks in from_one_core:
                    all_tokens.append(tokens)
                    merge_dicts_(all_unks, unks)
            return all_tokens, all_unks



import logging
import logging.config
import os

import appdirs
import yaml
import matplotlib
matplotlib.use('PS')  # to avoid error on OSX - this should go here, before other imports

from langmodels.nn import add_gru_to_model_data

app_name = 'langmodels'

current_script_location = os.path.realpath(__file__)
root_package_dir = os.path.dirname(current_script_location)
project_dir = os.path.dirname(root_package_dir)


def _get_version():
    with open(os.path.join(root_package_dir, 'VERSION')) as version_file:
        return version_file.read().strip()


def _get_major_version():
    return int(__version__.split('.')[0])


__version__ = _get_version()
__major_version__ = _get_major_version()

user_data_dir = appdirs.user_data_dir(app_name, appauthor=False, version=_get_version())
MODEL_ZOO_PATH = os.path.join(user_data_dir, 'modelzoo')


def load_logging_config():
    path = os.path.join(root_package_dir, 'logging.yaml')
    if os.path.exists(path):
        with open(path, 'rt') as f:
            logging_config = yaml.safe_load(f.read())
        logging.config.dictConfig(logging_config)
    else:
        logging.basicConfig(level=logging.DEBUG)

load_logging_config()

logging.getLogger('matplotlib').setLevel(logging.INFO)
logging.getLogger('everett').setLevel(logging.INFO)
logging.getLogger('urllib3.connectionpool').setLevel(logging.INFO)

add_gru_to_model_data()



import logging
from dataclasses import dataclass
from pprint import pprint, pformat

from fastai.basic_data import DataBunch
from fastai.layers import FlattenedLoss, CrossEntropyFlat
from fastai.text import LMLabelList, Vocab, TextList
from pathlib import Path
from typing import Sequence

from langmodels.profiling import get_cpu_memory_used_mb
from langmodels.tensor_ops import contains_no_value
from langmodels.training.numericalize import Numericalizer

logger = logging.getLogger(__name__)


UNKNOWN_TOKEN_INDEX = 0


@dataclass
class EmptyDataBunch(object):
    class EmptyTrainDS(object):
        y: LMLabelList = None

    class EmptyTrainDL(object):
        dataset = None

        def __len__(self):
            return 1
    vocab: Vocab
    path: str
    device: str
    backwards: bool = False
    loss_func: FlattenedLoss = CrossEntropyFlat()
    train_ds: EmptyTrainDS = EmptyTrainDS()
    train_dl: EmptyTrainDL = EmptyTrainDL()


def create_databunch(path_to_prep_dataset: str, file_paths: Sequence[Path], vocab: Vocab,
                     bs: int, bptt: int, device: str,
                     only_validation_files: bool = False, allow_unks: bool = False, verbose: bool = True) -> DataBunch:
    if verbose:
        logger.info(f'Getting preprocessed corpus from {path_to_prep_dataset}')
    numericalizer = Numericalizer(vocab, allow_unks=allow_unks, large_databunch=verbose)
    text_list = TextList(file_paths, path=path_to_prep_dataset, processor=numericalizer)

    if verbose:
        logger.info("Splitting into training/validation sets")
    if only_validation_files:
        split_list = text_list.split_by_valid_func(lambda f: True)
    else:
        split_list = text_list.split_by_folder()

    if verbose:
        logger.info("Labeling for langmodeling")
    labelled_list = split_list.label_for_lm()

    if verbose:
        cpu_memory_used_mb = get_cpu_memory_used_mb()
        logger.debug(f"Cpu memory used: {cpu_memory_used_mb} MB")

    if verbose:
        logger.info("Creating data bunches")
    data_bunched = labelled_list.databunch(bs=bs, bptt=bptt, device=device)
    return data_bunched


def check_data(data_bunch: DataBunch, vocab: Vocab, verbose: bool) -> None:
    first_batch = data_bunch.one_batch()[0]

    if not contains_no_value(first_batch, UNKNOWN_TOKEN_INDEX):
        raise ValueError(f"Unknown is found : {[vocab.textify(seq) for seq in first_batch]}")
    if verbose:
        logger.info(f'Displaying the first batch:\n{first_batch}')
        token_seqs = [vocab.textify(seq) for seq in first_batch]
        logger.info(pformat(token_seqs))



import os

import torch

from langmodels import project_dir
from langmodels.modelregistry import load_from_path


def test_hidden_state_not_changed_after_beam_search():
    trained_model = load_from_path(os.path.join(project_dir, 'data/models/dev_10k_1_10_190923.132328'))
    trained_model.feed_text("Hi there, it's")
    before = trained_model.model[0].hidden
    trained_model.predict_next_full_token(n_suggestions=10)
    after = trained_model.model[0].hidden

    for aa,bb in zip(after, before):
        for a, b in zip(aa, bb):
            assert torch.equal(b, a)


import os

from setuptools import setup, find_packages

root_package_name = 'langmodels'


def readme():
    with open('README.md') as f:
        return f.read()


def version():
    with open(os.path.join(root_package_name, 'VERSION')) as version_file:
        return version_file.read().strip()


with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()


setup(name='giganticode-langmodels',
      version=version(),
      description='A toolkit for applying machine learning to large source code corpora',
      long_description=readme(),
      long_description_content_type="text/markdown",
      url='http://github.com/giganticode/langmodels',
      author='giganticode',
      author_email='hlibbabii@gmail.com',
      license='MIT',
      packages=find_packages(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.7',
          'Operating System :: POSIX :: Linux',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Topic :: Scientific/Engineering :: Artificial Intelligence'
      ],
      python_requires='>=3.6',
      keywords='big large data source code corpus machine learning nlp '
               'pytorch torch fastai language modeling',
      install_requires=requirements,
      include_package_data=True,
      zip_safe=False)



import logging
import os

from comet_ml import Experiment
from fastai.basic_train import LearnerCallback, Learner
from fastai.callback import Callback
from fastai.callbacks import SaveModelCallback
from retrying import retry
from typing import Any

from langmodels.lmconfig.datamodel import ExperimentRun

logger = logging.getLogger(__name__)


def change_path_to_permanent(experiment_run: ExperimentRun, learner: Learner) -> None:
    old_full_path: str = str(learner.path / learner.model_dir)
    new_full_path: str = experiment_run.perm_path_to_model
    new_path, new_model_dir = os.path.split(new_full_path)
    if str(learner.path) != new_path:
        raise AssertionError(f"When changing model path to permanent, path to models must remain the same. "
                             f"However trying to rename: {old_full_path} -> {new_full_path}")
    learner.model_dir = new_model_dir
    os.rename(old_full_path, new_full_path)


def report_successful_experiment_to_comet(experiment: Experiment):
    experiment.log_parameter("model_available", True)


class FirstModelTrainedCallback(LearnerCallback):
    def __init__(self, learner: Learner, experiment_run: ExperimentRun):
        super().__init__(learner)
        self.experiment_run = experiment_run

    def on_epoch_end(self, epoch: int, **kwargs: Any) -> None:
        if epoch == 0:
            self.experiment_run.set_first_model_trained()
            change_path_to_permanent(self.experiment_run, super().learn)
            if self.experiment_run.comet_experiment:
                report_successful_experiment_to_comet(self.experiment_run.comet_experiment)


class LrLogger(Callback):
    def __init__(self, learner, experiment: Experiment = None):
        super().__init__()
        self.learn = learner
        self.experiment = experiment

    def on_batch_end(self, **kwargs):
        num_batch = kwargs['num_batch']
        epoch = kwargs['epoch']

        self.experiment.log_metric('lr', self.learn.opt.lr, step=num_batch, epoch=epoch)


class RetryingSaveModelCalback(SaveModelCallback):
    def __init__(self, learn: Learner, monitor: str = 'valid_loss', mode: str = 'auto', every: str = 'improvement',
                 name: str = 'bestmodel'):
        super().__init__(learn, monitor, mode, every, name)

    def _retry_if_io_error(exception):
        """Return True if we should retry (in this case when it's an IOError), False otherwise"""
        exception_type_to_retry = IOError
        wil_retry =  isinstance(exception, exception_type_to_retry)
        if wil_retry:
            logger.warning(f"Caught exception: {repr(exception)}. Will retry after some time ...")
        else:
            logger.warning(f"Exception was raised: {repr(exception)}. Retrying only {exception_type_to_retry}")
        return wil_retry

    WAIT_MULTIPLIER = 30 * 1000
    WAIT_MAX = 3600 * 1000

    @retry(retry_on_exception=_retry_if_io_error, wait_exponential_multiplier=WAIT_MULTIPLIER, wait_exponential_max=WAIT_MAX)
    def on_epoch_end(self, epoch: int, **kwargs: Any) -> None:
        super().on_epoch_end(epoch, **kwargs)

    @retry(retry_on_exception=_retry_if_io_error, wait_exponential_multiplier=WAIT_MULTIPLIER, wait_exponential_max=WAIT_MAX)
    def on_train_end(self, **kwargs):
        super().on_train_end(**kwargs)